# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
import copy
import distutils.util
import itertools
import logging
import os
import pathlib
import re
import shutil
import stat
import urllib
from typing import Any, Dict, List, Optional, Set  # noqa

import requests

from snapcraft import extractors, file_utils, formatting_utils, shell_utils, yaml_utils
from snapcraft.extractors import _metadata
from snapcraft.internal import common, errors, project_loader, states
from snapcraft.internal.deprecations import handle_deprecation_notice
from snapcraft.internal.meta import _manifest, _version
from snapcraft.internal.meta import errors as meta_errors
from snapcraft.internal.meta.application import ApplicationAdapter
from snapcraft.internal.meta.snap import Snap
from snapcraft.internal.project_loader import _config
from snapcraft.project import _schema

logger = logging.getLogger(__name__)


# From snapd's snap/validate.go, appContentWhitelist, with a slight modification: don't
# allow leading slashes.
_APP_COMMAND_PATTERN = re.compile("^[A-Za-z0-9. _#:$-][A-Za-z0-9/. _#:$-]*$")


def create_snap_packaging(project_config: _config.Config) -> str:
    """Create snap.yaml and related assets in meta.

    Create the meta directory and provision it with snap.yaml in the snap dir
    using information from config_data. Also copy in the local 'snap'
    directory, and generate wrappers for hooks coming from parts.

    :param dict config_data: project values defined in snapcraft.yaml.
    :return: meta_dir.
    """

    # Update config_data using metadata extracted from the project
    extracted_metadata = _update_yaml_with_extracted_metadata(
        project_config.data, project_config.parts, project_config.project.prime_dir
    )

    # Now that we've updated config_data with random stuff extracted from
    # parts, re-validate it to ensure the it still conforms with the schema.
    validator = _schema.Validator(project_config.data)
    validator.validate(source="properties")

    # Get the required grade if the global state file exists.
    if os.path.exists(project_config.project._get_global_state_file_path()):
        required_grade = states.GlobalState.load(
            filepath=project_config.project._get_global_state_file_path()
        ).get_required_grade()
    else:
        required_grade = None

    # Update default values
    _update_yaml_with_defaults(
        project_config.data, project_config.validator.schema, required_grade
    )

    packaging = _SnapPackaging(project_config, extracted_metadata)
    packaging.cleanup()
    packaging.validate_common_ids()
    packaging.finalize_snap_meta_commands()
    packaging.finalize_snap_meta_command_chains()
    packaging.finalize_snap_meta_version()
    packaging.write_snap_yaml()
    packaging.setup_assets()
    packaging.generate_hook_wrappers()
    packaging.write_snap_directory()
    packaging.warn_ld_library_paths()

    return packaging.meta_dir


def _update_yaml_with_extracted_metadata(
    config_data: Dict[str, Any],
    parts_config: project_loader.PartsConfig,
    prime_dir: str,
) -> Optional[extractors.ExtractedMetadata]:
    if "adopt-info" not in config_data:
        return None

    part_name = config_data["adopt-info"]
    part = parts_config.get_part(part_name)
    if not part:
        raise meta_errors.AdoptedPartMissingError(part_name)

    pull_state = part.get_pull_state()
    build_state = part.get_build_state()
    stage_state = part.get_stage_state()
    prime_state = part.get_prime_state()

    # Get the metadata from the pull step first.
    metadata = pull_state.extracted_metadata["metadata"]

    # Now update it using the metadata from the build step (i.e. the data
    # from the build step takes precedence over the pull step).
    metadata.update(build_state.extracted_metadata["metadata"])

    # Now make sure any scriptlet data are taken into account. Later steps
    # take precedence, and scriptlet data (even in earlier steps) take
    # precedence over extracted data.
    metadata.update(pull_state.scriptlet_metadata)
    metadata.update(build_state.scriptlet_metadata)
    metadata.update(stage_state.scriptlet_metadata)
    metadata.update(prime_state.scriptlet_metadata)

    if not metadata:
        # If we didn't end up with any metadata, let's ensure this part was
        # actually supposed to parse info. If not, let's try to be very
        # clear about what's happening, here. We do this after checking for
        # metadata because metadata could be supplied by scriptlets, too.
        if "parse-info" not in config_data["parts"][part_name]:
            raise meta_errors.AdoptedPartNotParsingInfo(part_name)

    _adopt_info(config_data, metadata, prime_dir)

    return metadata


def _adopt_info(
    config_data: Dict[str, Any],
    extracted_metadata: _metadata.ExtractedMetadata,
    prime_dir: str,
):
    ignored_keys = _adopt_keys(config_data, extracted_metadata, prime_dir)
    if ignored_keys:
        logger.warning(
            "The {keys} {plural_property} {plural_is} specified in adopted "
            "info as well as the YAML: taking the {plural_property} from the "
            "YAML".format(
                keys=formatting_utils.humanize_list(list(ignored_keys), "and"),
                plural_property=formatting_utils.pluralize(
                    ignored_keys, "property", "properties"
                ),
                plural_is=formatting_utils.pluralize(ignored_keys, "is", "are"),
            )
        )


def _adopt_keys(
    config_data: Dict[str, Any],
    extracted_metadata: _metadata.ExtractedMetadata,
    prime_dir: str,
) -> Set[str]:
    ignored_keys = set()
    metadata_dict = extracted_metadata.to_dict()

    # desktop_file_paths and common_ids are special cases that will be handled
    # after all the top level snapcraft.yaml keys.
    ignore = ("desktop_file_paths", "common_id")
    overrides = ((k, v) for k, v in metadata_dict.items() if k not in ignore)

    for key, value in overrides:
        if key in config_data:
            ignored_keys.add(key)
        else:
            if key == "icon":
                # Extracted appstream icon paths will be relative to prime.
                icon = pathlib.Path(prime_dir, str(value))
                if not icon.exists():
                    # Cannot find icon, ignore silently.
                    continue

                if _find_icon_file() is None:
                    # Already have icon file, do not overwrite.
                    continue

            config_data[key] = value

    if "desktop_file_paths" in metadata_dict and "common_id" in metadata_dict:
        app_name = _get_app_name_from_common_id(
            config_data, str(metadata_dict["common_id"])
        )
        if app_name and not _desktop_file_exists(app_name):
            for desktop_file_path in [
                os.path.join(prime_dir, d) for d in metadata_dict["desktop_file_paths"]
            ]:
                if os.path.exists(desktop_file_path):
                    config_data["apps"][app_name]["desktop"] = desktop_file_path
                    break

    return ignored_keys


def _find_icon_file() -> Optional[pathlib.Path]:
    """Check if the icon is specified as a file in the assets dir.

    The icon file can be in two directories: 'setup/gui' (deprecated) or
    'snap/gui'. The icon can be named 'icon.png' or 'icon.svg'.

    :returns: Path of found icon, None otherwise.
    """
    for icon_path in (
        pathlib.Path(asset_dir, "gui", icon_file)
        for (asset_dir, icon_file) in itertools.product(
            ["setup", "snap"], ["icon.png", "icon.svg"]
        )
    ):
        if icon_path.exists():
            return icon_path

    return None


def _get_app_name_from_common_id(
    config_data: Dict[str, Any], common_id: str
) -> Optional[str]:
    """Get the snap app name with a common-id.

    :params dict config_data: Project values defined in snapcraft.yaml.
    :params common_id: The common identifier across multiple packaging
        formats.
    :returns: The name of the snap app with the common-id.

    """
    if "apps" in config_data:
        for app in config_data["apps"]:
            if config_data["apps"][app].get("common-id") == common_id:
                return app
    return None


def _desktop_file_exists(app_name: str) -> bool:
    """Check if the desktop file is specified as a file in the assets dir.

    The desktop file can be in two directories: 'setup/gui' (deprecated) or
    'snap/gui'.

    :params app_name: The name of the snap app.
    :returns: True if the desktop file exists, False otherwise.
    """
    for desktop_path in (
        os.path.join(asset_dir, "gui", "{}.desktop".format(app_name))
        for asset_dir in ["setup", "snap"]
    ):
        if os.path.exists(desktop_path):
            return True
    else:
        return False


def _update_yaml_with_defaults(
    config_data, schema, required_grade: Optional[str]
) -> None:
    # Ensure that grade and confinement have their defaults applied, if
    # necessary. Defaults are taken from the schema. Technically these are the
    # only two optional keywords currently WITH defaults, but we don't want to
    # risk setting something that we add later on accident.
    if "confinement" not in config_data:
        default = schema["confinement"]["default"]
        config_data["confinement"] = default
        logger.warning(
            f"'confinement' property not specified: defaulting to {default!r}"
        )

    # Set the grade (or validate).
    if "grade" not in config_data:
        if required_grade is None:
            grade = schema["grade"]["default"]
        else:
            grade = required_grade
        logger.warning(f"'grade' property not specified: defaulting to {grade!r}.")
        config_data["grade"] = grade
    elif config_data["grade"] == "stable" and required_grade == "devel":
        raise meta_errors.GradeDevelRequiredError(set_grade=config_data["grade"])

    # Set default adapter
    app_schema = schema["apps"]["patternProperties"]["^[a-zA-Z0-9](?:-?[a-zA-Z0-9])*$"][
        "properties"
    ]
    default_adapter = app_schema["adapter"]["default"]
    for app in config_data.get("apps", {}).values():
        if "adapter" not in app and "command-chain" not in app:
            app["adapter"] = default_adapter
        elif "adapter" not in app and "command-chain" in app:
            app["adapter"] = "full"


def _check_passthrough_duplicates_section(yaml: Dict[str, Any]) -> None:
    # Any value already in the original dictionary must
    # not be specified in passthrough at the same time.
    if "passthrough" not in yaml:
        return

    passthrough = yaml["passthrough"]
    duplicates = list(yaml.keys() & passthrough.keys())
    if duplicates:
        raise meta_errors.AmbiguousPassthroughKeyError(duplicates)


def _check_passthrough_duplicates(snap_yaml: Dict[str, Any]) -> None:
    """Check passthrough properties for duplicate keys.

    Passthrough options may override assumed/extracted configuration,
    so this check must be done on the original YAML."""

    for section in ["apps", "hooks"]:
        if section not in snap_yaml:
            continue

        for value in snap_yaml[section].values():
            _check_passthrough_duplicates_section(value)

    _check_passthrough_duplicates_section(snap_yaml)


class _SnapPackaging:
    def __init__(
        self,
        project_config: _config.Config,
        extracted_metadata: Optional[_metadata.ExtractedMetadata],
    ) -> None:
        self._project_config = project_config
        self._extracted_metadata = extracted_metadata
        self._snapcraft_yaml_path = project_config.project.info.snapcraft_yaml_file_path
        self._prime_dir = project_config.project.prime_dir
        self._parts_dir = project_config.project.parts_dir

        self._arch_triplet = project_config.project.arch_triplet
        self.meta_dir = os.path.join(self._prime_dir, "meta")
        self.meta_gui_dir = os.path.join(self.meta_dir, "gui")
        self._config_data = project_config.data.copy()
        self._original_snapcraft_yaml = project_config.project.info.get_raw_snapcraft()

        self._install_path_pattern = re.compile(
            r"{}/[a-z0-9][a-z0-9+-]*/install".format(re.escape(self._parts_dir))
        )

        os.makedirs(self.meta_dir, exist_ok=True)

        # TODO: create_snap_packaging managles config data, so we create
        # a new private instance of snap_meta.  Longer term, this needs
        # to converge with project's snap_meta.
        self._snap_meta = Snap.from_dict(project_config.data)

    def cleanup(self):
        if os.path.exists(self.meta_gui_dir):
            for f in os.listdir(self.meta_gui_dir):
                if os.path.splitext(f)[1] == ".desktop":
                    os.remove(os.path.join(self.meta_gui_dir, f))

    def _finalize_icon(self) -> Optional[pathlib.Path]:
        """Ensure sure icon is properly configured and installed.

        Fetch from a remote URL, if required, and place in the meta/gui
        directory.
        """

        # Nothing to do if no icon is configured, search for existing icon.
        icon: Optional[str] = self._config_data.get("icon")
        if icon is None and self._extracted_metadata is not None:
            icon = self._extracted_metadata.get_icon()

        if icon is None:
            return _find_icon_file()

        # Extracted appstream icon paths will either:
        # (1) point to a file relative to prime
        # (2) point to a remote http(s) url
        #
        # The 'icon' specified in the snapcraft.yaml has the same
        # constraint as (2) and would have already been validated
        # as existing by the schema.  So we can treat it the same
        # at this point, regardless of the source of the icon.
        parsed_url = urllib.parse.urlparse(icon)
        parsed_path = pathlib.Path(parsed_url.path)
        icon_ext = parsed_path.suffix[1:]
        target_icon_path = pathlib.Path(self.meta_gui_dir, f"icon.{icon_ext}")

        target_icon_path.parent.mkdir(parents=True, exist_ok=True)
        if parsed_url.scheme in ["http", "https"]:
            # Remote - fetch URL and write to target.
            logger.info(f"Fetching icon from {icon!r}.")
            icon_data = requests.get(icon).content
            target_icon_path.write_bytes(icon_data)
        elif parsed_url.scheme == "":
            source_path = pathlib.Path(
                self._prime_dir,
                parsed_path.relative_to("/")
                if parsed_path.is_absolute()
                else parsed_path,
            )
            if source_path.exists():
                # Local with path relative to prime.
                file_utils.link_or_copy(str(source_path), str(target_icon_path))
            elif parsed_path.exists():
                # Local with path relative to project.
                file_utils.link_or_copy(str(parsed_path), str(target_icon_path))
            else:
                # No icon found, fall back to searching for existing icon.
                return _find_icon_file()
        else:
            raise RuntimeError(f"Unexpected icon path: {parsed_url!r}")

        return target_icon_path

    def finalize_snap_meta_commands(self) -> None:
        for app_name, app in self._snap_meta.apps.items():
            # Prime commands only if adapter != "none",
            # otherwise leave as-is.
            if app.adapter != ApplicationAdapter.NONE:
                app.prime_commands(
                    base=self._project_config.project.info.base,
                    prime_dir=self._prime_dir,
                )

    def finalize_snap_meta_command_chains(self) -> None:
        snapcraft_runner = self._generate_snapcraft_runner()
        if snapcraft_runner is None:
            return

        for app_name, app in self._snap_meta.apps.items():
            # Add runner to command chain if adapter is not "none".
            if app.adapter != ApplicationAdapter.NONE:
                app.command_chain.insert(0, snapcraft_runner)

    def finalize_snap_meta_version(self) -> None:
        # Reparse the version, the order should stick.
        version = self._config_data["version"]
        version_script = self._config_data.get("version-script")

        if version_script:
            # Deprecation warning for use of version-script.
            handle_deprecation_notice("dn10")

        self._snap_meta.version = _version.get_version(version, version_script)

    def write_snap_yaml(self) -> None:
        # Ensure snap meta is valid before writing.
        self._snap_meta.validate()
        _check_passthrough_duplicates(self._original_snapcraft_yaml)

        package_snap_path = os.path.join(self.meta_dir, "snap.yaml")
        self._snap_meta.write_snap_yaml(path=package_snap_path)

    def warn_ld_library_paths(self) -> None:
        root_ld_library_path = self._snap_meta.environment.get("LD_LIBRARY_PATH")
        # Dictionary of app names with LD_LIBRARY_PATH in their environment.
        app_environment: Dict[str, str] = dict()

        for app_name, app_props in self._config_data.get("apps", dict()).items():
            with contextlib.suppress(KeyError):
                app_environment[app_name] = app_props["environment"]["LD_LIBRARY_PATH"]

        if root_ld_library_path is None and not app_environment:
            return

        ld_library_path_empty: Set[str] = set()
        if root_ld_library_path is None and app_environment:
            ld_library_path_empty = {
                name
                for name, ld_env in app_environment.items()
                if "$LD_LIBRARY_PATH" in ld_env or "${LD_LIBRARY_PATH}" in ld_env
            }
        elif (
            root_ld_library_path is not None
            and "LD_LIBRARY_PATH" in root_ld_library_path
        ):
            ld_library_path_empty = {"."}

        _EMPTY_LD_LIBRARY_PATH_ITEM_PATTERN = re.compile("^:|::|:$")

        for name, ld_env in app_environment.items():
            if _EMPTY_LD_LIBRARY_PATH_ITEM_PATTERN.findall(ld_env):
                ld_library_path_empty.add(name)

        if (
            root_ld_library_path is not None
            and _EMPTY_LD_LIBRARY_PATH_ITEM_PATTERN.findall(root_ld_library_path)
        ):
            ld_library_path_empty.add(".")

        if ld_library_path_empty:
            logger.warning(
                "CVE-2020-27348: A potentially empty LD_LIBRARY_PATH has been set for environment "
                "in {}. "
                "The current working directory will be added to the library path if empty. "
                "This can cause unexpected libraries to be loaded.".format(
                    formatting_utils.humanize_list(sorted(ld_library_path_empty), "and")
                )
            )

    def setup_assets(self) -> None:
        # We do _setup_from_setup first since it is legacy and let the
        # declarative items take over.
        self._setup_gui()

        icon_path = self._finalize_icon()
        if icon_path is not None:
            if str(icon_path).startswith(self._prime_dir):
                icon_path = icon_path.relative_to(self._prime_dir)
            relative_icon_path: Optional[str] = str(icon_path)
        else:
            relative_icon_path = None

        snap_name = self._project_config.project.info.name
        for app_name, app in self._snap_meta.apps.items():
            app.write_command_wrappers(prime_dir=self._prime_dir)
            app.write_application_desktop_file(
                snap_name=snap_name,
                prime_dir=self._prime_dir,
                gui_dir=self.meta_gui_dir,
                icon_path=relative_icon_path,
            )
            app.validate_command_chain_executables(self._prime_dir)

        if self._config_data.get("type", "") == "gadget":
            if not os.path.exists("gadget.yaml"):
                raise errors.MissingGadgetError()
            file_utils.link_or_copy(
                "gadget.yaml", os.path.join(self.meta_dir, "gadget.yaml")
            )
        if self._config_data.get("type", "") == "kernel":
            if os.path.exists("kernel.yaml"):
                file_utils.link_or_copy(
                    "kernel.yaml", os.path.join(self.meta_dir, "kernel.yaml")
                )

    def _assemble_runtime_environment(self) -> str:
        # Classic confinement or building on a host that does not match the target base
        # means we cannot setup an environment that will work.
        if self._config_data["confinement"] == "classic":
            # Temporary workaround for snapd bug not expanding PATH:
            # We generate an empty runner which addresses the issue.
            # https://bugs.launchpad.net/snapd/+bug/1860369
            return ""

        env = list()
        if self._project_config.project._snap_meta.base in ("core", "core16", "core18"):
            common.env = self._project_config.snap_env()
            assembled_env = common.assemble_env()
            common.reset_env()

            assembled_env = assembled_env.replace(self._prime_dir, "$SNAP")
            env.append(self._install_path_pattern.sub("$SNAP", assembled_env))
        else:
            # TODO use something local to the meta package and
            # only add paths for directory items that actually exist.
            runtime_env = project_loader.runtime_env(
                self._prime_dir, self._project_config.project.arch_triplet
            )
            for e in runtime_env:
                env.append(re.sub(self._prime_dir, "$SNAP", e))

        if all(
            [
                part._build_attributes.enable_patchelf()
                for part in self._project_config.all_parts
            ]
        ):
            # All ELF files have had rpath and interpreter patched. Strip all LD_LIBRARY_PATH variables
            env = [e for e in env if not e.startswith("export LD_LIBRARY_PATH=")]
        else:
            env.append(
                'export LD_LIBRARY_PATH="$SNAP_LIBRARY_PATH${LD_LIBRARY_PATH:+:$LD_LIBRARY_PATH}"'
            )

        return "\n".join(env)

    def _generate_snapcraft_runner(self) -> Optional[str]:
        """Create runner if required.

        Return path relative to prime directory, if created."""

        # If there are no apps, or type is snapd, no need to create a runner.
        if not self._snap_meta.apps or self._config_data.get("type") == "snapd":
            return None

        # No more command-chain for core20 and classic confinement.
        # This was a workaround for LP: #1860369.
        if (
            self._snap_meta.base not in ("core", "core16", "core18", None)
            and self._snap_meta.confinement == "classic"
        ):
            return None

        assembled_env = self._assemble_runtime_environment()

        meta_runner = os.path.join(
            self._prime_dir, "snap", "command-chain", "snapcraft-runner"
        )

        os.makedirs(os.path.dirname(meta_runner), exist_ok=True)
        with open(meta_runner, "w") as f:
            print("#!/bin/sh", file=f)
            print(assembled_env, file=f)
            print('exec "$@"', file=f)
        os.chmod(meta_runner, 0o755)

        return os.path.relpath(meta_runner, self._prime_dir)

    def _record_manifest_and_source_snapcraft_yaml(self):
        prime_snap_dir = os.path.join(self._prime_dir, "snap")
        recorded_snapcraft_yaml_path = os.path.join(prime_snap_dir, "snapcraft.yaml")
        if os.path.isfile(recorded_snapcraft_yaml_path):
            os.unlink(recorded_snapcraft_yaml_path)
        manifest_file_path = os.path.join(prime_snap_dir, "manifest.yaml")
        if os.path.isfile(manifest_file_path):
            os.unlink(manifest_file_path)

        # FIXME hide this functionality behind a feature flag for now
        if distutils.util.strtobool(os.environ.get("SNAPCRAFT_BUILD_INFO", "n")):
            os.makedirs(prime_snap_dir, exist_ok=True)
            shutil.copy2(self._snapcraft_yaml_path, recorded_snapcraft_yaml_path)
            annotated_snapcraft = _manifest.annotate_snapcraft(
                self._project_config.project, copy.deepcopy(self._config_data)
            )
            with open(manifest_file_path, "w") as manifest_file:
                yaml_utils.dump(annotated_snapcraft, stream=manifest_file)

    def write_snap_directory(self) -> None:
        """Record manifest and copy assets found under $SNAPCRAFT_PROJECT_ROOT/snap.

        These assets have priority over any code generated assets and include:
        - hooks
        - gui
        """
        snap_assets_dir = self._project_config.project._get_snapcraft_assets_dir()
        prime_snap_dir = os.path.join(self._prime_dir, "snap")

        snap_dir_iter = itertools.product([prime_snap_dir], ["hooks"])
        meta_dir_iter = itertools.product([self.meta_dir], ["hooks", "gui"])

        for origin in itertools.chain(snap_dir_iter, meta_dir_iter):
            src_dir = os.path.join(snap_assets_dir, origin[1])
            dst_dir = os.path.join(origin[0], origin[1])
            if os.path.isdir(src_dir):
                os.makedirs(dst_dir, exist_ok=True)
                for asset in os.listdir(src_dir):
                    source = os.path.join(src_dir, asset)
                    destination = os.path.join(dst_dir, asset)

                    with contextlib.suppress(FileNotFoundError):
                        os.remove(destination)

                    file_utils.link_or_copy(source, destination, follow_symlinks=True)

                    # Ensure that the hook is executable in meta/hooks, this is a moot
                    # point considering the prior link_or_copy call, but is technically
                    # correct and allows for this operation to take place only once.
                    if origin[0] == self.meta_dir and origin[1] == "hooks":
                        _prepare_hook(destination)

        self._record_manifest_and_source_snapcraft_yaml()

    def generate_hook_wrappers(self) -> None:
        """Setup hooks in meta.

        Hooks generated with this criteria:

        - A stub for every hook definition that contains a command-chain entry
          to ensure the command-chain for a defined hook runs. A command-chain
          with no hook can occur when using extensions.
        - A wrapper from a hook created from hooks found under
          self._prime_dir/snap/hooks, these hooks are usually generated by parts.
        """
        hooks_dir = os.path.join(self._prime_dir, "meta", "hooks")
        snap_hooks_dir = os.path.join(self._prime_dir, "snap", "hooks")
        hooks_in_snap_dir = (
            os.listdir(snap_hooks_dir) if os.path.isdir(snap_hooks_dir) else []
        )
        hooks_with_command_chain = [
            hook
            for hook in self._snap_meta.hooks.values()
            if hook.command_chain is not None
            and hook.hook_name not in hooks_in_snap_dir
        ]

        if hooks_with_command_chain or hooks_in_snap_dir:
            os.makedirs(hooks_dir, exist_ok=True)

        # Create stub hooks as necessary.
        for hook in hooks_with_command_chain:
            hook_path = pathlib.Path(hooks_dir) / hook.hook_name
            if not hook_path.exists():
                hook_path.write_text("#!/bin/sh\n")
                hook_path.chmod(0o755)

        # Write wrapper hooks as necessary.
        for hook_name in hooks_in_snap_dir:
            file_path = os.path.join(snap_hooks_dir, hook_name)
            # Make sure the hook is executable
            _prepare_hook(file_path)

            hook_exec = os.path.join("$SNAP", "snap", "hooks", hook_name)
            hook_path = pathlib.Path(hooks_dir) / hook_name
            with contextlib.suppress(FileNotFoundError):
                hook_path.unlink()

            self._write_wrap_exe(hook_exec, str(hook_path))

    def _setup_gui(self):
        # Handles the setup directory which only contains gui assets.
        setup_dir = "setup"
        if not os.path.exists(setup_dir):
            return

        handle_deprecation_notice("dn3")

        gui_src = os.path.join(setup_dir, "gui")
        if os.path.exists(gui_src):
            for f in os.listdir(gui_src):
                if not os.path.exists(self.meta_gui_dir):
                    os.mkdir(self.meta_gui_dir)
                shutil.copy2(os.path.join(gui_src, f), self.meta_gui_dir)

    def _write_wrap_exe(self, wrapexec, wrappath, shebang=None, args=None, cwd=None):
        assembled_env = self._assemble_runtime_environment()

        if args:
            quoted_args = ['"{}"'.format(arg) for arg in args]
        else:
            quoted_args = []
        args = " ".join(quoted_args) + ' "$@"' if args else '"$@"'
        cwd = "cd {}".format(cwd) if cwd else ""

        executable = '"{}"'.format(wrapexec)

        if shebang:
            if shebang.startswith("/usr/bin/env "):
                shebang = shell_utils.which(shebang.split()[1])
            new_shebang = self._install_path_pattern.sub("$SNAP", shebang)
            new_shebang = re.sub(self._prime_dir, "$SNAP", new_shebang)
            if new_shebang != shebang:
                # If the shebang was pointing to and executable within the
                # local 'parts' dir, have the wrapper script execute it
                # directly, since we can't use $SNAP in the shebang itself.
                executable = '"{}" "{}"'.format(new_shebang, wrapexec)

        with open(wrappath, "w+") as f:
            print("#!/bin/sh", file=f)
            if cwd:
                print("{}".format(cwd), file=f)
            print(assembled_env, file=f)
            print("exec {} {}".format(executable, args), file=f)

        os.chmod(wrappath, 0o755)

    def validate_common_ids(self) -> None:
        if (
            not self._extracted_metadata
            or not self._extracted_metadata.common_id_list
            or "apps" not in self._config_data
        ):
            return

        common_id_list = self._extracted_metadata.common_id_list
        for app in self._config_data["apps"]:
            app_common_id = self._config_data["apps"][app].get("common-id")
            if app_common_id not in common_id_list:
                logger.warning(
                    "Common ID {common_id!r} specified in app {app!r} is "
                    "not used in any metadata file.".format(
                        common_id=app_common_id, app=app
                    )
                )


def _prepare_hook(hook_path):
    # Ensure hook is executable
    if not os.stat(hook_path).st_mode & stat.S_IEXEC:
        os.chmod(hook_path, 0o755)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from testtools.matchers import Contains, Equals, Not

from snapcraft.internal import log
from tests import fixture_setup, unit


class LogTestCase(unit.TestCase):
    def setUp(self):
        super().setUp()

        self.info_color = log._ColoredFormatter.LEVEL_COLORS["INFO"]
        self.warning_color = log._ColoredFormatter.LEVEL_COLORS["WARNING"]
        self.error_color = log._ColoredFormatter.LEVEL_COLORS["ERROR"]
        self.critical_color = log._ColoredFormatter.LEVEL_COLORS["CRITICAL"]

    def test_configure_must_send_messages_to_stdout(self):
        logger_name = self.id()
        log.configure(logger_name)
        logger = logging.getLogger(logger_name)
        # Overwrite the level to log everything.
        logger.setLevel(logging.DEBUG)

        logger.debug("Test debug")
        logger.info("Test info")

        stdout = self.fake_terminal.getvalue()
        self.assertThat(stdout, Contains("Test debug"))
        expected_info = "{}Test info\033[0m".format(self.info_color)
        self.assertThat(stdout, Contains(expected_info))
        self.assertThat(self.fake_terminal.getvalue(stderr=True), Equals(""))

    def test_configure_must_send_warnings_and_errors_to_stderr(self):
        logger_name = self.id()
        log.configure(logger_name)
        logger = logging.getLogger(logger_name)
        # Overwrite the level to log everything.
        logger.setLevel(logging.DEBUG)

        logger.warning("Test warning")
        logger.error("Test error")
        logger.critical("Test critical")

        stderr = self.fake_terminal.getvalue(stderr=True)
        expected_warning = "{}Test warning\033[0m".format(self.warning_color)
        self.assertThat(stderr, Contains(expected_warning))
        expected_error = "{}Test error\033[0m".format(self.error_color)
        self.assertThat(stderr, Contains(expected_error))
        expected_crit = "{}Test critical\033[0m".format(self.critical_color)
        self.assertThat(stderr, Contains(expected_crit))
        self.assertThat(self.fake_terminal.getvalue(), Equals(""))

    def test_configure_must_log_info_and_higher(self):
        logger_name = self.id()
        log.configure(logger_name)
        logger = logging.getLogger(logger_name)

        logger.debug("Test debug")
        logger.info("Test info")
        logger.warning("Test warning")
        logger.error("Test error")
        logger.critical("Test critical")

        stdout = self.fake_terminal.getvalue()
        self.assertThat(stdout, Not(Contains("Test debug")))
        expected_info = "{}Test info\033[0m".format(self.info_color)
        self.assertThat(stdout, Contains(expected_info))

        stderr = self.fake_terminal.getvalue(stderr=True)
        expected_warning = "{}Test warning\033[0m".format(self.warning_color)
        self.assertThat(stderr, Contains(expected_warning))
        expected_error = "{}Test error\033[0m".format(self.error_color)
        self.assertThat(stderr, Contains(expected_error))
        expected_crit = "{}Test critical\033[0m".format(self.critical_color)
        self.assertThat(stderr, Contains(expected_crit))

    def test_configure_must_support_debug(self):
        logger_name = self.id()
        log.configure(logger_name, log_level=logging.DEBUG)
        logger = logging.getLogger(logger_name)

        logger.debug("Test debug")
        logger.info("Test info")
        logger.warning("Test warning")
        logger.error("Test error")
        logger.critical("Test critical")

        stdout = self.fake_terminal.getvalue()
        self.assertThat(stdout, Contains("Test debug"))
        expected_info = "{}Test info\033[0m".format(self.info_color)
        self.assertThat(stdout, Contains(expected_info))

        stderr = self.fake_terminal.getvalue(stderr=True)
        expected_warning = "{}Test warning\033[0m".format(self.warning_color)
        self.assertThat(stderr, Contains(expected_warning))
        expected_error = "{}Test error\033[0m".format(self.error_color)
        self.assertThat(stderr, Contains(expected_error))
        expected_crit = "{}Test critical\033[0m".format(self.critical_color)
        self.assertThat(stderr, Contains(expected_crit))

    def test_configure_must_support_no_tty(self):
        self.fake_terminal = fixture_setup.FakeTerminal(isatty=False)
        self.useFixture(self.fake_terminal)
        logger_name = self.id()
        log.configure(logger_name, log_level=logging.DEBUG)
        logger = logging.getLogger(logger_name)

        logger.debug("Test debug")
        logger.info("Test info")
        logger.warning("Test warning")
        logger.error("Test error")
        logger.critical("Test critical")

        stdout = self.fake_terminal.getvalue()
        self.assertThat(stdout, Contains("Test debug"))
        self.assertThat(stdout, Contains("Test info"))
        self.assertThat(stdout, Not(Contains(self.info_color)))

        stderr = self.fake_terminal.getvalue(stderr=True)
        self.assertThat(stderr, Contains("Test warning"))
        self.assertThat(stderr, Not(Contains(self.warning_color)))
        self.assertThat(stderr, Not(Contains("\033[0m")))
        self.assertThat(stderr, Contains("Test error"))
        self.assertThat(stderr, Not(Contains(self.error_color)))
        self.assertThat(stderr, Contains("Test critical"))
        self.assertThat(stderr, Not(Contains(self.critical_color)))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import distutils.util
import os
import sys
from typing import Any, Dict, List, Optional

import click

from snapcraft.cli.echo import confirm, prompt, warning
from snapcraft.internal import common, errors
from snapcraft.internal.meta.snap import Snap
from snapcraft.project import Project, get_snapcraft_yaml


class PromptOption(click.Option):
    def prompt_for_value(self, ctx):
        default = self.get_default(ctx)

        # If this is a prompt for a flag we need to handle this
        # differently.
        if self.is_bool_flag:
            return confirm(self.prompt, default)

        return prompt(
            self.prompt,
            default=default,
            type=self.type,
            hide_input=self.hide_input,
            show_choices=self.show_choices,
            confirmation_prompt=self.confirmation_prompt,
            value_proc=lambda x: self.process_value(ctx, x),
        )


class BoolParamType(click.ParamType):
    name = "boolean"

    def convert(self, value, param, ctx):
        """Convert option string to value.

        Unlike click's BoolParamType, use distutils.util.strtobool to
        convert values.
        """
        if isinstance(value, bool):
            return value
        try:
            return bool(distutils.util.strtobool(value))
        except ValueError:
            self.fail("%r is not a valid boolean" % value, param, ctx)

    def __repr__(self):
        return "BOOL"


_SUPPORTED_PROVIDERS = ["host", "lxd", "multipass"]
_HIDDEN_PROVIDERS = ["managed-host"]
_ALL_PROVIDERS = _SUPPORTED_PROVIDERS + _HIDDEN_PROVIDERS
_PROVIDER_OPTIONS: List[Dict[str, Any]] = [
    dict(
        param_decls="--target-arch",
        metavar="<arch>",
        help="Target architecture to cross compile to",
        supported_providers=["host", "lxd", "multipass"],
    ),
    dict(
        param_decls="--debug",
        is_flag=True,
        help="Shells into the environment if the build fails.",
        supported_providers=["host", "lxd", "managed-host", "multipass"],
    ),
    dict(
        param_decls="--shell",
        is_flag=True,
        help="Shells into the environment in lieu of the step to run.",
        supported_providers=["host", "lxd", "managed-host", "multipass"],
    ),
    dict(
        param_decls="--shell-after",
        is_flag=True,
        help="Shells into the environment after the step has run.",
        supported_providers=["host", "lxd", "managed-host", "multipass"],
    ),
    dict(
        param_decls="--destructive-mode",
        is_flag=True,
        help="Forces snapcraft to try and use the current host to build (implies `--provider=host`).",
        supported_providers=["host", "managed-host"],
    ),
    dict(
        param_decls="--use-lxd",
        is_flag=True,
        help="Forces snapcraft to use LXD to build (implies `--provider=lxd`).",
        supported_providers=["lxd"],
    ),
    dict(
        param_decls="--provider",
        envvar="SNAPCRAFT_BUILD_ENVIRONMENT",
        show_envvar=False,
        help="Build provider to use.",
        metavar="[{}]".format("|".join(_SUPPORTED_PROVIDERS)),
        type=click.Choice(_ALL_PROVIDERS),
        supported_providers=_ALL_PROVIDERS,
    ),
    dict(
        param_decls="--http-proxy",
        metavar="<http-proxy>",
        help="HTTP proxy for host build environments.",
        envvar="http_proxy",
        supported_providers=["host", "lxd", "managed-host", "multipass"],
    ),
    dict(
        param_decls="--https-proxy",
        metavar="<https-proxy>",
        help="HTTPS proxy for host build environments.",
        envvar="https_proxy",
        supported_providers=["host", "lxd", "managed-host", "multipass"],
    ),
    dict(
        param_decls="--add-ca-certificates",
        metavar="<certificate-path>",
        help="File or directory containing CA certificates to install into build environments.",
        envvar="SNAPCRAFT_ADD_CA_CERTIFICATES",
        supported_providers=["lxd", "multipass"],
        type=click.Path(
            exists=True, file_okay=True, dir_okay=True, readable=True, resolve_path=True
        ),
    ),
    dict(
        param_decls="--bind-ssh",
        is_flag=True,
        help="Bind ~/.ssh directory to locally-run build environments.",
        envvar="SNAPCRAFT_BIND_SSH",
        supported_providers=["lxd", "multipass"],
    ),
    dict(
        param_decls="--enable-developer-debug",
        is_flag=True,
        help="Enable developer debug logging.",
        envvar="SNAPCRAFT_ENABLE_DEVELOPER_DEBUG",
        supported_providers=["host", "lxd", "managed-host", "multipass"],
        hidden=True,
    ),
    dict(
        param_decls="--enable-manifest",
        is_flag=True,
        type=BoolParamType(),
        help="Generate snap manifest.",
        envvar="SNAPCRAFT_BUILD_INFO",
        supported_providers=["host", "lxd", "managed-host", "multipass"],
        hidden=True,
    ),
    dict(
        param_decls="--manifest-image-information",
        metavar="<json string>",
        help="Set snap manifest image-info",
        envvar="SNAPCRAFT_IMAGE_INFO",
        supported_providers=["host", "lxd", "managed-host", "multipass"],
        hidden=True,
    ),
    dict(
        param_decls="--enable-experimental-extensions",
        is_flag=True,
        help="Enable extensions that are experimental and not considered stable.",
        envvar="SNAPCRAFT_ENABLE_EXPERIMENTAL_EXTENSIONS",
        supported_providers=["host", "lxd", "managed-host", "multipass"],
    ),
    dict(
        param_decls="--enable-experimental-target-arch",
        is_flag=True,
        help="Enable experimental `--target-arch` support for core20.",
        envvar="SNAPCRAFT_ENABLE_EXPERIMENTAL_TARGET_ARCH",
        supported_providers=["host", "lxd", "managed-host", "multipass"],
    ),
    dict(
        param_decls="--ua-token",
        metavar="<ua-token>",
        help="Configure build environment with ESM using specified UA token.",
        envvar="SNAPCRAFT_UA_TOKEN",
        supported_providers=["host", "lxd", "managed-host", "multipass"],
    ),
]


def _add_options(options, func, hidden):
    for option in reversed(options):
        # Pop param_decls option to prevent exception further on down the
        # line for: `got multiple values for keyword argument`.
        option = option.copy()
        param_decls = option.pop("param_decls")

        # Pop supported_providers option because it is snapcraft-only.
        if "supported_providers" in option:
            option.pop("supported_providers")

        hidden_override = option.pop("hidden", hidden)
        click_option = click.option(param_decls, **option, hidden=hidden_override)
        func = click_option(func)
    return func


def add_provider_options(hidden=False):
    def _add_provider_options(func):
        return _add_options(_PROVIDER_OPTIONS, func, hidden)

    return _add_provider_options


def _sanity_check_build_provider_flags(build_provider: str, **kwargs) -> None:
    destructive_mode = kwargs.get("destructive_mode")
    env_provider = os.getenv("SNAPCRAFT_BUILD_ENVIRONMENT")

    # Specifying --provider=host requires the use of --destructive-mode.
    # Exceptions include:
    # (1) SNAPCRAFT_BUILD_ENVIRONMENT=host.
    # (2) Running inside of a container.
    if (
        build_provider == "host"
        and not env_provider == "host"
        and not destructive_mode
        and not common.is_process_container()
    ):
        raise click.BadArgumentUsage(
            "--provider=host requires --destructive-mode to acknowledge side effects"
        )

    if env_provider and env_provider != build_provider:
        raise click.BadArgumentUsage(
            "mismatch between --provider={} and SNAPCRAFT_BUILD_ENVIRONMENT={}".format(
                build_provider, env_provider
            )
        )

    # Error if any sys.argv params are for unsupported providers.
    # Values from environment variables and configuration files only
    # change defaults, so they are safe to ignore due to filtering
    # in get_build_provider_flags().
    for option in _PROVIDER_OPTIONS:
        key: str = option["param_decls"]  # type: ignore
        supported_providers: List[str] = option["supported_providers"]  # type: ignore
        if key in sys.argv and build_provider not in supported_providers:
            raise click.BadArgumentUsage(
                f"{key} cannot be used with build provider {build_provider!r}"
            )

    # Check if running as sudo but only if the host is not managed-host where Snapcraft
    # runs as root already. This effectively avoids the warning when using the default
    # build provider (Multipass) that uses "sudo" to get "root".
    if (
        build_provider != "managed-host"
        and os.getenv("SUDO_USER")
        and os.geteuid() == 0
    ):
        warning(
            "Running with 'sudo' may cause permission errors and is discouraged. Use 'sudo' when cleaning."
        )


def get_build_provider(skip_sanity_checks: bool = False, **kwargs) -> str:
    """Get build provider and determine if running as managed instance."""

    provider = kwargs.get("provider")

    if not provider:
        if kwargs.get("use_lxd"):
            provider = "lxd"
        elif kwargs.get("destructive_mode"):
            provider = "host"
        elif common.is_process_container():
            provider = "host"
        else:
            # Default is multipass.
            provider = "multipass"

    # Sanity checks may be skipped for the purpose of checking legacy.
    if not skip_sanity_checks:
        _sanity_check_build_provider_flags(provider, **kwargs)

    return provider


def _param_decls_to_kwarg(key: str) -> str:
    """Format a param_decls to keyword argument name."""

    # Drop leading "--".
    key = key.replace("--", "", 1)

    # Convert dashes to underscores.
    return key.replace("-", "_")


def get_build_provider_flags(build_provider: str, **kwargs) -> Dict[str, str]:
    """Get configured options applicable to build_provider."""

    build_provider_flags: Dict[str, str] = dict()

    # Should not happen - developer safety check.
    if build_provider not in _ALL_PROVIDERS:
        raise RuntimeError(f"Invalid build provider: {build_provider}")

    for option in _PROVIDER_OPTIONS:
        key: str = option["param_decls"]  # type: ignore
        is_flag: bool = option.get("is_flag", False)  # type: ignore
        envvar: Optional[str] = option.get("envvar")  # type: ignore
        supported_providers: List[str] = option["supported_providers"]  # type: ignore

        # Skip --provider option.
        if key == "--provider":
            continue

        # Skip options that do not apply to configured provider.
        if build_provider not in supported_providers:
            continue

        # Skip boolean flags that have not been set.
        key_formatted = _param_decls_to_kwarg(key)
        if is_flag and not kwargs.get(key_formatted):
            continue

        # Add build provider flag using envvar as key.
        if envvar is not None and key_formatted in kwargs:
            build_provider_flags[envvar] = kwargs[key_formatted]

    return build_provider_flags


def get_project(*, is_managed_host: bool = False, **kwargs):
    # We need to do this here until we can get_snapcraft_yaml as part of Project.
    if is_managed_host:
        try:
            os.chdir(os.path.expanduser(os.path.join("~", "project")))
        except FileNotFoundError:
            # No project found (fresh environment).
            raise errors.ProjectNotFoundError()

    snapcraft_yaml_file_path = get_snapcraft_yaml()

    # This method may be called from a click.Command with no parent.
    ctx = click.get_current_context()
    if ctx.parent is not None:
        for key, value in ctx.parent.params.items():
            if not kwargs.get(key):
                kwargs[key] = value

    project = Project(
        debug=kwargs.pop("debug", False),
        target_deb_arch=kwargs.pop("target_arch", None),
        snapcraft_yaml_file_path=snapcraft_yaml_file_path,
        is_managed_host=is_managed_host,
    )

    # Validate yaml info from schema prior to consumption.
    if project.info is not None and not project.is_legacy_project():
        project.info.validate_raw_snapcraft()

        # TODO: this should be automatic on get_project().
        # This is not the complete meta parsed by the project loader.
        project._snap_meta = Snap.from_dict(project.info.get_raw_snapcraft())

    return project


def apply_host_provider_flags(build_provider_flags: Dict[str, str]) -> None:
    """Set build environment flags in snapcraft process."""

    # Snapcraft plugins currently check for environment variables,
    # e.g. http_proxy and https_proxy.  Ensure they are set if configured.
    for key, value in build_provider_flags.items():
        if value is None:
            if key in os.environ:
                os.environ.pop(key)
        else:
            os.environ[key] = str(value)

    # Clear false/unset boolean environment flags.
    for option in _PROVIDER_OPTIONS:
        if not option.get("is_flag", False):
            continue

        env_name = option.get("envvar")
        if env_name is None:
            continue

        if not build_provider_flags.get(env_name):
            os.environ.pop(env_name, None)
            continue

    issue_build_provider_warnings(build_provider_flags)


def issue_build_provider_warnings(build_provider_flags: Dict[str, str]) -> None:
    # Log any experimental flags in use.
    if build_provider_flags.get("SNAPCRAFT_ENABLE_EXPERIMENTAL_EXTENSIONS"):
        warning("*EXPERIMENTAL* extensions enabled.")

    if build_provider_flags.get("SNAPCRAFT_ENABLE_EXPERIMENTAL_TARGET_ARCH"):
        warning("*EXPERIMENTAL* --target-arch for core20 enabled.")



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import pathlib
from unittest.mock import MagicMock, call, patch

from testtools.matchers import Equals

from snapcraft import file_utils
from snapcraft.internal import errors, pluginhandler, steps
from tests.unit import TestCase, load_part


class CleanTestCase(TestCase):
    @patch.object(pluginhandler.PluginHandler, "is_clean")
    @patch("os.rmdir")
    @patch("os.listdir")
    def test_clean_part_that_exists(self, mock_listdir, mock_rmdir, mock_is_clean):
        mock_listdir.return_value = False
        mock_is_clean.return_value = True

        part_name = "test_part"
        partdir = os.path.join(self.parts_dir, part_name)
        os.makedirs(partdir)

        p = self.load_part(part_name)
        p.clean()

        mock_listdir.assert_called_once_with(partdir)
        mock_rmdir.assert_called_once_with(partdir)

    @patch("os.rmdir")
    @patch("os.listdir")
    @patch("os.path.exists")
    def test_clean_part_already_clean(self, mock_exists, mock_listdir, mock_rmdir):
        mock_exists.return_value = False

        part_name = "test_part"
        p = self.load_part(part_name)
        p.clean()

        partdir = os.path.join(self.parts_dir, part_name)
        mock_exists.assert_has_calls([call(partdir)])
        self.assertFalse(mock_listdir.called)
        self.assertFalse(mock_rmdir.called)

    @patch.object(pluginhandler.PluginHandler, "is_clean")
    @patch("os.rmdir")
    @patch("os.listdir")
    def test_clean_part_remaining_parts(self, mock_listdir, mock_rmdir, mock_is_clean):
        mock_listdir.return_value = True
        mock_is_clean.return_value = True

        part_name = "test_part"
        partdir = os.path.join(self.parts_dir, part_name)
        os.makedirs(partdir)

        p = self.load_part(part_name)
        p.clean()

        mock_listdir.assert_called_once_with(partdir)
        self.assertFalse(mock_rmdir.called)

    def test_clean_prime_multiple_independent_parts(self):
        # Create part1 and get it through the "build" step.
        handler1 = self.load_part("part1")
        handler1.makedirs()

        bindir = os.path.join(handler1.part_install_dir, "bin")
        os.makedirs(bindir)
        open(os.path.join(bindir, "1"), "w").close()

        handler1.mark_done(steps.BUILD)

        # Now create part2 and get it through the "build" step.
        handler2 = self.load_part("part2")
        handler2.makedirs()

        bindir = os.path.join(handler2.part_install_dir, "bin")
        os.makedirs(bindir)
        open(os.path.join(bindir, "2"), "w").close()

        handler2.mark_done(steps.BUILD)

        # Now stage both parts
        handler1.stage()
        handler2.stage()

        # And prime both parts
        handler1.prime()
        handler2.prime()

        # Verify that part1's file has been primeped
        self.assertTrue(os.path.exists(os.path.join(self.prime_dir, "bin", "1")))

        # Verify that part2's file has been primeped
        self.assertTrue(os.path.exists(os.path.join(self.prime_dir, "bin", "2")))

        # Now clean the prime step for part1
        handler1.clean_prime({})

        # Verify that part1's file is no longer primeped
        self.assertFalse(
            os.path.exists(os.path.join(self.prime_dir, "bin", "1")),
            "Expected part1's primeped files to be cleaned",
        )

        # Verify that part2's file is still there
        self.assertTrue(
            os.path.exists(os.path.join(self.prime_dir, "bin", "2")),
            "Expected part2's primeped files to be untouched",
        )

    def test_clean_prime_after_fileset_change(self):
        # Create part1 and get it through the "build" step.
        handler = self.load_part("part1")
        handler.makedirs()

        bindir = os.path.join(handler.part_install_dir, "bin")
        os.makedirs(bindir)
        open(os.path.join(bindir, "1"), "w").close()
        open(os.path.join(bindir, "2"), "w").close()

        handler.mark_done(steps.BUILD)
        handler.stage()
        handler.prime()

        # Verify that both files have been primeped
        self.assertTrue(os.path.exists(os.path.join(self.prime_dir, "bin", "1")))
        self.assertTrue(os.path.exists(os.path.join(self.prime_dir, "bin", "2")))

        # Now update the `snap` fileset to only snap one of these files
        handler.plugin.options.snap = ["bin/1"]

        # Now clean the prime step for part1
        handler.clean_prime({})

        # Verify that part1's file is no longer primeped
        self.assertFalse(
            os.path.exists(os.path.join(self.prime_dir, "bin", "1")),
            "Expected bin/1 to be cleaned",
        )
        self.assertFalse(
            os.path.exists(os.path.join(self.prime_dir, "bin", "2")),
            "Expected bin/2 to be cleaned as well, even though the filesets "
            "changed since it was primeped.",
        )

    def test_clean_old_prime_state(self):
        handler = self.load_part("test-part")
        handler.makedirs()

        open(os.path.join(self.prime_dir, "1"), "w").close()

        handler.mark_done(steps.PRIME, None)

        self.assertTrue(os.path.exists(handler.plugin.partdir))

        handler.clean()

        self.assertFalse(os.path.exists(handler.plugin.partdir))

    def test_clean_prime_old_prime_state(self):
        handler = self.load_part("test-part")
        handler.makedirs()

        primed_file = os.path.join(self.prime_dir, "1")
        open(primed_file, "w").close()

        handler.mark_done(steps.PRIME, None)

        raised = self.assertRaises(
            errors.MissingStateCleanError, handler.clean, step=steps.PRIME
        )

        self.assertThat(raised.step, Equals(steps.PRIME))
        self.assertTrue(os.path.isfile(primed_file))

    def test_clean_stage_multiple_independent_parts(self):
        # Create part1 and get it through the "build" step.
        handler1 = self.load_part("part1")
        handler1.makedirs()

        bindir = os.path.join(handler1.part_install_dir, "bin")
        os.makedirs(bindir)
        open(os.path.join(bindir, "1"), "w").close()

        handler1.mark_done(steps.BUILD)

        # Now create part2 and get it through the "build" step.
        handler2 = self.load_part("part2")
        handler2.makedirs()

        bindir = os.path.join(handler2.part_install_dir, "bin")
        os.makedirs(bindir)
        open(os.path.join(bindir, "2"), "w").close()

        handler2.mark_done(steps.BUILD)

        # Now stage both parts
        handler1.stage()
        handler2.stage()

        # Verify that part1's file has been staged
        self.assertTrue(os.path.exists(os.path.join(self.stage_dir, "bin", "1")))

        # Verify that part2's file has been staged
        self.assertTrue(os.path.exists(os.path.join(self.stage_dir, "bin", "2")))

        # Now clean the stage step for part1
        handler1.clean_stage({})

        # Verify that part1's file is no longer staged
        self.assertFalse(
            os.path.exists(os.path.join(self.stage_dir, "bin", "1")),
            "Expected part1's staged files to be cleaned",
        )

        # Verify that part2's file is still there
        self.assertTrue(
            os.path.exists(os.path.join(self.stage_dir, "bin", "2")),
            "Expected part2's staged files to be untouched",
        )

    def test_clean_stage_after_fileset_change(self):
        # Create part1 and get it through the "build" step.
        handler = self.load_part("part1")
        handler.makedirs()

        bindir = os.path.join(handler.part_install_dir, "bin")
        os.makedirs(bindir)
        open(os.path.join(bindir, "1"), "w").close()
        open(os.path.join(bindir, "2"), "w").close()

        handler.mark_done(steps.BUILD)
        handler.stage()

        # Verify that both files have been staged
        self.assertTrue(os.path.exists(os.path.join(self.stage_dir, "bin", "1")))
        self.assertTrue(os.path.exists(os.path.join(self.stage_dir, "bin", "2")))

        # Now update the `stage` fileset to only snap one of these files
        handler.plugin.options.stage = ["bin/1"]

        # Now clean the prime step for part1
        handler.clean_stage({})

        # Verify that part1's file is no longer staged
        self.assertFalse(
            os.path.exists(os.path.join(self.stage_dir, "bin", "1")),
            "Expected bin/1 to be cleaned",
        )
        self.assertFalse(
            os.path.exists(os.path.join(self.stage_dir, "bin", "2")),
            "Expected bin/2 to be cleaned as well, even though the filesets "
            "changed since it was staged.",
        )

    def test_clean_old_stage_state(self):
        handler = self.load_part("part1")
        handler.makedirs()

        open(os.path.join(self.stage_dir, "1"), "w").close()

        handler.mark_done(steps.STAGE, None)

        self.assertTrue(os.path.exists(handler.plugin.partdir))

        handler.clean()

        self.assertFalse(os.path.exists(handler.plugin.partdir))

    def test_clean_stage_old_stage_state(self):
        handler = self.load_part("test-part")
        handler.makedirs()

        staged_file = os.path.join(self.stage_dir, "1")
        open(staged_file, "w").close()

        handler.mark_done(steps.STAGE, None)

        raised = self.assertRaises(
            errors.MissingStateCleanError, handler.clean, step=steps.STAGE
        )

        self.assertThat(raised.step, Equals(steps.STAGE))
        self.assertTrue(os.path.isfile(staged_file))


class TestCleanPrime:

    scenarios = [
        ("all", {"fileset": ["*"]}),
        ("no1", {"fileset": ["-1"]}),
        ("onlya", {"fileset": ["a"]}),
        ("onlybase", {"fileset": ["*", "-*/*"]}),
        ("only1a", {"fileset": ["1/a"]}),
        ("nostara", {"fileset": ["-*/a"]}),
    ]

    def test_clean_prime(self, monkeypatch, tmp_work_path, fileset):
        monkeypatch.setattr(file_utils, "get_snap_tool_path", lambda x: x)

        handler = load_part("test_part", part_properties={"prime": fileset})
        handler.makedirs()

        installdir = pathlib.Path(handler.part_install_dir)

        (installdir / "1/1a/1b").mkdir(parents=True)
        (installdir / "2/2a").mkdir(parents=True)
        (installdir / "3").mkdir(parents=True)

        (installdir / "a").touch()
        (installdir / "b").touch()
        (installdir / "1/a").touch()
        (installdir / "3/a").touch()

        handler.mark_done(steps.BUILD)

        # Stage the installed files
        handler.stage()

        # Now prime them
        handler.prime()

        assert os.listdir(handler._project.prime_dir) != []

        handler.clean_prime({})

        assert os.listdir(handler._project.prime_dir) == []


class TestCleanStage:

    scenarios = [
        ("all", {"fileset": ["*"]}),
        ("no1", {"fileset": ["-1"]}),
        ("onlya", {"fileset": ["a"]}),
        ("onlybase", {"fileset": ["*", "-*/*"]}),
        ("only1a", {"fileset": ["1/a"]}),
        ("nostara", {"fileset": ["-*/a"]}),
    ]

    def test_clean_stage(self, tmp_work_path, fileset):
        handler = load_part("test_part", part_properties={"stage": fileset})
        handler.makedirs()

        installdir = pathlib.Path(handler.part_install_dir)

        (installdir / "1/1a/1b").mkdir(parents=True)
        (installdir / "2/2a").mkdir(parents=True)
        (installdir / "3").mkdir(parents=True)

        (installdir / "a").touch()
        (installdir / "b").touch()
        (installdir / "1/a").touch()
        (installdir / "3/a").touch()

        handler.mark_done(steps.BUILD)

        # Stage the installed files
        handler.stage()

        assert os.listdir(handler._project.stage_dir) != []

        handler.clean_stage({})

        assert os.listdir(handler._project.stage_dir) == []


class PerStepCleanTestCase(TestCase):
    def setUp(self):
        super().setUp()

        self.manager_mock = MagicMock()

        patcher = patch.object(pluginhandler.PluginHandler, "clean_pull")
        self.manager_mock.attach_mock(patcher.start(), "clean_pull")
        self.addCleanup(patcher.stop)

        patcher = patch.object(pluginhandler.PluginHandler, "clean_build")
        self.manager_mock.attach_mock(patcher.start(), "clean_build")
        self.addCleanup(patcher.stop)

        patcher = patch.object(pluginhandler.PluginHandler, "clean_stage")
        self.manager_mock.attach_mock(patcher.start(), "clean_stage")
        self.addCleanup(patcher.stop)

        patcher = patch.object(pluginhandler.PluginHandler, "clean_prime")
        self.manager_mock.attach_mock(patcher.start(), "clean_prime")
        self.addCleanup(patcher.stop)

        self.handler = self.load_part("test_part")

    def test_clean_pull_order(self):
        self.handler.clean(step=steps.PULL)

        # Verify the step cleaning order
        self.assertThat(len(self.manager_mock.mock_calls), Equals(4))
        self.manager_mock.assert_has_calls(
            [
                call.clean_prime({}),
                call.clean_stage({}),
                call.clean_build(),
                call.clean_pull(),
            ]
        )

    def test_clean_build_order(self):
        self.handler.clean(step=steps.BUILD)

        # Verify the step cleaning order
        self.assertThat(len(self.manager_mock.mock_calls), Equals(3))
        self.manager_mock.assert_has_calls(
            [call.clean_prime({}), call.clean_stage({}), call.clean_build()]
        )

    def test_clean_stage_order(self):
        self.handler.clean(step=steps.STAGE)

        # Verify the step cleaning order
        self.assertThat(len(self.manager_mock.mock_calls), Equals(2))
        self.manager_mock.assert_has_calls([call.clean_prime({}), call.clean_stage({})])

    def test_clean_prime_order(self):
        self.handler.clean(step=steps.PRIME)

        # Verify the step cleaning order
        self.assertThat(len(self.manager_mock.mock_calls), Equals(1))
        self.manager_mock.assert_has_calls([call.clean_prime({})])


class CleanPullTestCase(TestCase):
    def test_clean_pull_directory(self):
        handler = self.load_part("test-part")

        handler.pull()
        source_file = os.path.join(handler.part_source_dir, "source")
        open(source_file, "w").close()

        handler.clean_pull()

        # The source directory should now be gone
        self.assertFalse(os.path.exists(handler.part_source_dir))

    def test_clean_pull_symlink(self):
        real_source_directory = os.path.join(os.getcwd(), "src")
        os.mkdir(real_source_directory)

        handler = self.load_part("test-part", part_properties={"source": "src"})

        handler.pull()
        os.rmdir(handler.part_source_dir)
        os.symlink(real_source_directory, handler.part_source_dir)

        handler.clean_pull()

        # The source symlink should now be gone, but the real source should
        # still be there.
        self.assertFalse(os.path.exists(handler.part_source_dir))
        self.assertTrue(os.path.isdir(real_source_directory))


def test_clean_build(tmp_work_path):
    handler = load_part("test-part")

    handler.build()

    source_file = pathlib.Path(handler.part_source_dir) / "source"
    build_basedir = pathlib.Path(handler.plugin.build_basedir)
    source_file.touch()
    (build_basedir / "built").touch()
    (build_basedir / "built").touch()

    handler.clean_build()

    # Make sure the source file hasn't been touched
    assert source_file.exists()

    # Make sure the build directory is gone
    assert not build_basedir.exists()

    # Make sure the install directory is gone
    assert not pathlib.Path(handler.plugin.installdir).exists()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The make plugin is useful for building make based parts.

Make based projects are projects that have a Makefile that drives the
build.

This plugin always runs 'make' followed by 'make install', except when
the 'artifacts' keyword is used.

This plugin uses the common plugin keywords as well as those for "sources".
For more information check the 'plugins' topic for the former and the
'sources' topic for the latter.

Additionally, this plugin uses the following plugin-specific keywords:

    - make-parameters
      (list of strings)
      Pass the given parameters to the make command.
"""

from typing import Any, Dict, List, Set

from snapcraft.plugins.v2 import PluginV2


class MakePlugin(PluginV2):
    @classmethod
    def get_schema(cls) -> Dict[str, Any]:
        return {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "type": "object",
            "additionalProperties": False,
            "properties": {
                "make-parameters": {
                    "type": "array",
                    "uniqueItems": True,
                    "items": {"type": "string"},
                    "default": [],
                }
            },
        }

    def get_build_snaps(self) -> Set[str]:
        return set()

    def get_build_packages(self) -> Set[str]:
        return {"gcc", "make"}

    def get_build_environment(self) -> Dict[str, str]:
        return dict()

    def _get_make_command(self, target: str = "") -> str:
        cmd = ["make", '-j"${SNAPCRAFT_PARALLEL_BUILD_COUNT}"']

        if target:
            cmd.append(target)

        cmd.extend(self.options.make_parameters)

        return " ".join(cmd)

    def get_build_commands(self) -> List[str]:
        return [
            self._get_make_command(),
            '{} DESTDIR="$SNAPCRAFT_PART_INSTALL"'.format(
                self._get_make_command(target="install")
            ),
        ]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2021 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pathlib

import pytest

from snapcraft.storeapi.http_clients import errors
from snapcraft.storeapi.http_clients._config import Config


class ConfigImpl(Config):
    def _get_section_name(self) -> str:
        return "test-section"

    def _get_config_path(self) -> pathlib.Path:
        return pathlib.Path("config.cfg")


@pytest.fixture
def conf(tmp_work_path):
    conf = ConfigImpl()
    yield conf


def test_non_existing_file_succeeds(conf):
    assert conf.parser.sections() == []
    assert conf.is_section_empty() is True


def test_existing_file(conf):
    conf.set("foo", "bar")
    conf.save()

    conf.load()

    assert conf.get("foo") == "bar"
    assert conf.is_section_empty() is False


def test_irrelevant_sections_are_ignored(conf):
    with conf._get_config_path().open("w") as config_file:
        print("[example.com]", file=config_file)
        print("foo=bar", file=config_file)

    conf.load()

    assert conf.get("foo") is None


def test_clear_preserver_other_sections(conf):
    with conf._get_config_path().open("w") as config_file:
        print("[keep_me]", file=config_file)
        print("foo=bar", file=config_file)

    conf.load()
    conf.set("bar", "baz")

    assert conf.get("bar") == "baz"

    conf.clear()
    conf.save()
    conf.load()

    assert conf.get("bar") is None
    assert conf.get("foo", "keep_me") == "bar"
    assert conf.is_section_empty() is True


def test_save_encoded(conf):
    conf.set("bar", "baz")
    conf.save(encode=True)
    conf.load()

    assert conf.get("bar") == "baz"
    with conf._get_config_path().open() as config_file:
        assert config_file.read() == "W3Rlc3Qtc2VjdGlvbl0KYmFyID0gYmF6Cgo=\n"


def test_save_encoded_other_config_file(conf):
    conf.set("bar", "baz")
    test_config_file = pathlib.Path("test-config")
    with test_config_file.open("w") as config_fd:
        conf.save(config_fd=config_fd, encode=True)
        config_fd.flush()

    with test_config_file.open() as config_file:
        assert config_file.read() == "W3Rlc3Qtc2VjdGlvbl0KYmFyID0gYmF6Cgo=\n"


def test_load_invalid_config(conf):
    test_config_file = pathlib.Path("test-config")
    with test_config_file.open("w") as config_fd:
        print("invalid config", file=config_fd)
        config_fd.flush()

    with test_config_file.open() as config_fd:
        with pytest.raises(errors.InvalidLoginConfig):
            conf.load(config_fd=config_fd)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017, 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Any, Dict, List, Optional, Set, Union

from snapcraft import yaml_utils


class ExtractedMetadata(yaml_utils.SnapcraftYAMLObject):
    """Collection of metadata extracted from a part."""

    yaml_tag = "!ExtractedMetadata"

    def __init__(
        self,
        *,
        common_id: Optional[str] = None,
        title: Optional[str] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        version: Optional[str] = None,
        grade: Optional[str] = None,
        icon: Optional[str] = None,
        desktop_file_paths: Optional[List[str]] = None
    ) -> None:
        """Create a new ExtractedMetadata instance.

        :param str: common_id: The common identifier across multiple packaging
            formats
        :param str title: Extracted title
        :param str summary: Extracted summary
        :param str description: Extracted description
        :param str version: Extracted version
        :param str grade: Extracted grade
        :param str icon: Extracted icon
        :param list desktop_file_paths: Extracted desktop file paths
        """  # noqa

        self._data = {}  # type: Dict[str, Union[str, List[str]]]
        self.common_id_list = []  # type: List[str]

        if common_id:
            self._data["common_id"] = common_id
        if title:
            self._data["title"] = title
        if summary:
            self._data["summary"] = summary
        if description:
            self._data["description"] = description
        if version:
            self._data["version"] = version
        if grade:
            self._data["grade"] = grade
        if icon:
            self._data["icon"] = icon
        if desktop_file_paths:
            self._data["desktop_file_paths"] = desktop_file_paths

    def update(self, other: "ExtractedMetadata") -> None:
        """Update this metadata with other metadata.

        Note that the other metadata will take precedence, and may overwrite
        data contained here.

        :param ExtractedMetadata other: Metadata from which to update
        """
        self._data.update(other.to_dict())
        common_id = self.get_common_id()
        if common_id:
            self.common_id_list.append(common_id)

    def get_common_id(self) -> Optional[str]:
        """Return extracted common_id.

        :returns: Extracted common_id
        :rtype: str
        """
        common_id = self._data.get("common_id")
        return str(common_id) if common_id else None

    def get_title(self) -> Optional[str]:
        """Return extracted title.

        :returns: Extracted title
        :rtype: str
        """
        title = self._data.get("title")
        return str(title) if title else None

    def get_summary(self) -> Optional[str]:
        """Return extracted summary.

        :returns: Extracted summary
        :rtype: str
        """
        summary = self._data.get("summary")
        return str(summary) if summary else None

    def get_description(self) -> Optional[str]:
        """Return extracted description.

        :returns: Extracted description
        :rtype: str
        """
        description = self._data.get("description")
        return str(description) if description else None

    def get_version(self) -> Optional[str]:
        """Return extracted version.

        :returns: Extracted version
        :rtype: str
        """
        version = self._data.get("version")
        return str(version) if version else None

    def get_grade(self) -> Optional[str]:
        """Return extracted grade.

        :returns: Extracted grade
        :rtype: str
        """
        grade = self._data.get("grade")
        return str(grade) if grade else None

    def get_icon(self) -> Optional[str]:
        """Return extracted icon.

        :returns: Extracted icon
        :rtype: str
        """
        icon = self._data.get("icon")
        return str(icon) if icon else None

    def get_desktop_file_paths(self) -> Optional[List[str]]:
        """Return extracted desktop files paths.

        :returns: Extracted desktop files paths
        :rtype: list
        """
        desktop_file_paths = self._data.get("desktop_file_paths")
        return list(desktop_file_paths) if desktop_file_paths else None

    def to_dict(self) -> Dict[str, Union[str, List[str]]]:
        """Return all extracted metadata.

        :returns: All extracted metadata in dict form.
        :rtype: dict
        """
        return self._data.copy()

    def overlap(self, other: "ExtractedMetadata") -> Set[str]:
        """Return all overlapping keys between this and other.

        :returns: All overlapping keys between this and other
        :rtype: set
        """
        return set(self._data.keys() & other.to_dict().keys())

    def __str__(self) -> str:
        return str(self._data)

    def __eq__(self, other: Any) -> bool:
        if type(other) is type(self):
            return self._data == other._data

        return False

    def __len__(self) -> int:
        return self._data.__len__()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import subprocess

from tests import integration


class StagePackageVersionTestCase(integration.TestCase):
    def test_stage_package_with_invalid_version_must_fail(self):
        self.copy_project_to_cwd("stage-packages-missing-dependency")
        self.set_stage_package_version(
            os.path.join("snap", "snapcraft.yaml"),
            part="part-with-stage-packages",
            package="haskell-doc",
            version="invalid",
        )
        error = self.assertRaises(
            subprocess.CalledProcessError, self.run_snapcraft, "pull"
        )
        self.assertIn(
            "Failed to fetch stage packages: "
            "Error downloading packages for part "
            "'part-with-stage-packages': "
            "The package 'haskell-doc=invalid' was not found.",
            str(error.output),
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
from typing import TYPE_CHECKING, Iterable, List, Optional

from . import typing
from .errors import UnsatisfiedStatementError

# Don't use circular imports unless type checking
if TYPE_CHECKING:
    from ._processor import GrammarProcessor  # noqa: F401

_SELECTOR_PATTERN = re.compile(r"\Aon\s+([^,\s](?:,?[^,]+)*)\Z")
_WHITESPACE_PATTERN = re.compile(r"\A.*\s.*\Z")


class Statement:
    """Base class for all grammar statements"""

    def __init__(
        self,
        *,
        body: Optional[typing.Grammar],
        processor: "GrammarProcessor",
        call_stack: Optional[typing.CallStack],
        check_primitives: bool = False
    ) -> None:
        """Create an Statement instance.

        :param list body: The body of the clause.
        :param GrammarProcessor process: GrammarProcessor to use for processing
                                         this statement.
        :param list call_stack: Call stack leading to this statement.
        :param bool check_primitives: Whether or not the primitives should be
                                      checked for validity as part of
                                      evaluating the elses.
        """
        if call_stack:
            self.__call_stack = call_stack
        else:
            self.__call_stack = []

        self._body = body
        self._processor = processor
        self._check_primitives = check_primitives
        self._else_bodies: List[Optional[typing.Grammar]] = []

        self.__processed_body: Optional[List[str]] = None
        self.__processed_else: Optional[List[str]] = None

    def add_else(self, else_body: Optional[typing.Grammar]) -> None:
        """Add an 'else' clause to the statement.

        :param list else_body: The body of an 'else' clause.

        The 'else' clauses will be processed in the order they are added.
        """
        self._else_bodies.append(else_body)

    def process(self) -> List[str]:
        """Process this statement.

        :return: Primitives as determined by evaluating the statement or its
                 else clauses.
        """
        if self._check():
            return self._process_body()
        else:
            return self._process_else()

    def _process_body(self) -> List[str]:
        """Process the main body of this statement.

        :return: Primitives as determined by processing the main body.
        """
        if self.__processed_body is None:
            self.__processed_body = self._processor.process(
                grammar=self._body, call_stack=self._call_stack(include_self=True)
            )

        return self.__processed_body

    def _process_else(self) -> List[str]:
        """Process the else clauses of this statement in order.

        :return: Primitives as determined by processing the else clauses.
        """
        if self.__processed_else is not None:
            return self.__processed_else

        self.__processed_else = list()
        for else_body in self._else_bodies:
            if not else_body:
                # Handle the 'else fail' case.
                raise UnsatisfiedStatementError(self)

            processed_else = self._processor.process(
                grammar=else_body, call_stack=self._call_stack()
            )
            if processed_else:
                self.__processed_else = processed_else
                if not self._check_primitives or self._validate_primitives(
                    processed_else
                ):
                    break

        return self.__processed_else

    def _validate_primitives(self, primitives: Iterable[str]) -> bool:
        """Ensure that all primitives are valid.

        :param primitives: Iterable container of primitives.

        :return: Whether or not all primitives are valid.
        :rtype: bool
        """
        for primitive in primitives:
            if not self._processor.checker(primitive):
                return False
        return True

    def _call_stack(self, *, include_self=False) -> List["Statement"]:
        """The call stack when processing this statement.

        :param bool include_self: Whether or not this statement should be
                                  included in the stack.

        :return: The call stack
        :rtype: list
        """
        if include_self:
            return self.__call_stack + [self]
        else:
            return self.__call_stack

    def __repr__(self):
        return "{!r}".format(self.__str__())

    def _check(self) -> bool:
        """Check if a statement main body should be processed.

        :return: True if main body should be processed, False if elses should
                 be processed.
        :rtype: bool
        """
        raise NotImplementedError("this must be implemented by child classes")

    def __eq__(self, other) -> bool:
        raise NotImplementedError("this must be implemented by child classes")

    def __str__(self) -> str:
        raise NotImplementedError("this must be implemented by child classes")



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The go plugin can be used for go projects.

The plugin supports Go Modules if a go.mod definition is found in the
root of the part's source the Go version used is 1.13 or greater.

If using 1.11 or 1.12, enabling through environment flags is required.
Read more about this at https://github.com/golang/go/wiki/Modules and
the implementation will only install the binary corresponding to the
main module.

If a go.mod file is not found, the plugin will "go get".

This plugin uses the common plugin keywords as well as those for "sources".
For more information check the 'plugins' topic for the former and the
'sources' topic for the latter.

Additionally, this plugin uses the following plugin-specific keywords:

    - go-channel:
      (string, default: latest/stable)
      The Snap Store channel to install go from. If set to an empty string,
      go will be installed using the system's traditional package manager.

    - go-packages:
      (list of strings)
      Go packages to fetch, these must be a "main" package. Dependencies
      are pulled in automatically by `go get`.
      Packages that are not "main" will not cause an error, but would
      not be useful either.
      If the package is a part of the go-importpath the local package
      corresponding to those sources will be used.

    - go-importpath:
      (string)
      This entry tells the checked out `source` to live within a certain path
      within `GOPATH`.
      This entry is not required if `source` uses `go.mod`.

    - go-buildtags:
      (list of strings)
      Tags to use during the go build. Default is not to use any build tags.
"""

import fileinput
import logging
import os
import re
import shutil
from glob import iglob
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple

from pkg_resources import parse_version

from snapcraft import common
from snapcraft.internal import elf, errors
from snapcraft.plugins.v1 import PluginV1

if TYPE_CHECKING:
    from snapcraft.project import Project


logger = logging.getLogger(__name__)
# The absolute minimum required go version (including one with flags set).
_GO_MOD_REQUIRED_GO_VERSION = "1.11"
# Go versions lower than this require the flag.
_GO_MOD_ENV_FLAG_REQUIRED_GO_VERSION = "1.13"


class GoModRequiredVersionError(errors.SnapcraftException):
    def __init__(self, *, go_version: str) -> None:
        self._go_version = go_version

    def get_brief(self) -> str:
        return "Use of go.mod requires Go {!r} or greater, found: {!r}".format(
            _GO_MOD_REQUIRED_GO_VERSION, self._go_version
        )

    def get_resolution(self) -> str:
        return "Set go-channel to a newer version of Go and try again."


def _get_cgo_ldflags(library_paths: List[str]) -> str:
    cgo_ldflags: List[str] = list()

    existing_cgo_ldflags = os.getenv("CGO_LDFLAGS")
    if existing_cgo_ldflags:
        cgo_ldflags.append(existing_cgo_ldflags)

    flags = common.combine_paths(library_paths, "-L", " ")
    if flags:
        cgo_ldflags.append(flags)

    ldflags = os.getenv("LDFLAGS")
    if ldflags:
        cgo_ldflags.append(ldflags)

    return " ".join(cgo_ldflags)


class GoPlugin(PluginV1):
    @classmethod
    def schema(cls) -> Dict[str, Any]:
        schema = super().schema()
        schema["properties"]["go-channel"] = {
            "type": "string",
            "default": "latest/stable",
        }
        schema["properties"]["go-packages"] = {
            "type": "array",
            "minitems": 1,
            "uniqueItems": True,
            "items": {"type": "string"},
            "default": [],
        }
        schema["properties"]["go-importpath"] = {"type": "string", "default": ""}
        schema["properties"]["go-buildtags"] = {
            "type": "array",
            "minitems": 1,
            "uniqueItems": True,
            "items": {"type": "string"},
            "default": [],
        }
        schema["anyOf"] = [{"required": ["source"]}, {"required": ["go-packages"]}]

        return schema

    @classmethod
    def get_build_properties(cls) -> List[str]:
        # Inform Snapcraft of the properties associated with building. If these
        # change in the YAML Snapcraft will consider the build step dirty.
        return ["go-packages", "go-buildtags", "go-channel"]

    @classmethod
    def get_pull_properties(cls) -> List[str]:
        # Inform Snapcraft of the properties associated with pulling. If these
        # change in the YAML Snapcraft will consider the pull step dirty.
        return ["go-packages", "go-channel"]

    def __init__(self, name: str, options, project: "Project") -> None:
        super().__init__(name, options, project)

        self._setup_base_tools(options.go_channel, project._get_build_base())
        self._is_classic = project._snap_meta.confinement == "classic"

        self._install_bin_dir = os.path.join(self.installdir, "bin")

        self._gopath = os.path.join(self.partdir, "go")
        self._gopath_src = os.path.join(self._gopath, "src")
        self._gopath_bin = os.path.join(self._gopath, "bin")
        self._gopath_pkg = os.path.join(self._gopath, "pkg")

        self._version_regex = re.compile(r"^go version go(.*) .*$")
        self._go_version: Optional[Tuple[str, ...]] = None

        self._go_mod_warning_issued = False

    def _setup_base_tools(self, go_channel: str, base: Optional[str]) -> None:
        if go_channel:
            self.build_snaps.append("go/{}".format(go_channel))
        elif base in ("core", "core16", "core18"):
            self.build_packages.append("golang-go")
        else:
            raise errors.PluginBaseError(part_name=self.name, base=base)

    def _get_parsed_go_version(self) -> Tuple[str, ...]:
        if self._go_version is not None:
            return self._go_version

        go_version_cmd_output: str = self._run_output(["go", "version"])
        version_match = self._version_regex.match(go_version_cmd_output)

        if version_match is None:
            raise RuntimeError(
                "Unable to parse go version output: {!r}".format(go_version_cmd_output)
            )

        self._go_version = parse_version(version_match.group(1))

        return self._go_version

    def _is_using_go_mod(self, cwd: str) -> bool:
        if not os.path.exists(os.path.join(cwd, "go.mod")):
            return False

        current_version = self._get_parsed_go_version()

        if current_version < parse_version(_GO_MOD_REQUIRED_GO_VERSION):
            raise GoModRequiredVersionError(go_version=str(self._go_version))

        if not self._go_mod_warning_issued and current_version < parse_version(
            _GO_MOD_ENV_FLAG_REQUIRED_GO_VERSION
        ):
            logger.warning(
                "Ensure build environment configuration is correct for this version of Go. "
                "Read more about it at "
                "https://github.com/golang/go/wiki/Modules#how-to-install-and-activate-module-support"
            )
            self._go_mod_warning_issued = True

        return True

    def _pull_go_mod(self) -> None:
        self._run(["go", "mod", "download"], cwd=self.sourcedir)

    def _pull_go_packages(self) -> None:
        os.makedirs(self._gopath_src, exist_ok=True)

        # use -d to only download (build will happen later)
        # use -t to also get the test-deps
        # since we are not using -u the sources will stick to the
        # original checkout.
        if any(iglob("{}/**/*.go".format(self.sourcedir), recursive=True)):
            go_package = self._get_local_go_package()
            go_package_path = os.path.join(self._gopath_src, go_package)
            if os.path.islink(go_package_path):
                os.unlink(go_package_path)
            os.makedirs(os.path.dirname(go_package_path), exist_ok=True)
            os.symlink(self.sourcedir, go_package_path)
            self._run(["go", "get", "-t", "-d", "./{}/...".format(go_package)])

        for go_package in self.options.go_packages:
            self._run(["go", "get", "-t", "-d", go_package])

    def pull(self) -> None:
        super().pull()

        if self._is_using_go_mod(cwd=self.sourcedir):
            return self._pull_go_mod()
        else:
            self._pull_go_packages()

    def clean_pull(self) -> None:
        super().clean_pull()

        # Remove the gopath (if present)
        if os.path.exists(self._gopath):
            shutil.rmtree(self._gopath)

    def _get_local_go_package(self) -> str:
        if self.options.go_importpath:
            go_package = self.options.go_importpath
        else:
            logger.warning(
                "Please consider setting `go-importpath` for the {!r} "
                "part".format(self.name)
            )
            go_package = os.path.basename(os.path.abspath(self.options.source))
        return go_package

    def _get_local_main_packages(self) -> List[str]:
        search_path = "./{}/...".format(self._get_local_go_package())
        packages = self._run_output(
            ["go", "list", "-f", "{{.ImportPath}} {{.Name}}", search_path]
        )
        packages_split = [p.split() for p in packages.splitlines()]
        main_packages = [p[0] for p in packages_split if p[1] == "main"]
        return main_packages

    def _get_module(self) -> str:
        pattern = re.compile(r"module\s*(?P<module>.*)\s")
        with fileinput.input(os.path.join(self.builddir, "go.mod")) as go_mod_file:
            for line in go_mod_file:
                match = pattern.search(line)
                if match is not None:
                    module = match.group("module")
                    break
            else:
                raise RuntimeError("Expected a module in go.mod")

        return os.path.basename(module)

    def _build(self, *, package: str = "") -> None:
        build_cmd = ["go", "build"]

        if self.options.go_buildtags:
            build_cmd.extend(["-tags={}".format(",".join(self.options.go_buildtags))])

        relink_cmd = build_cmd + ["-ldflags", "-linkmode=external"]

        if self._is_using_go_mod(self.builddir) and not package:
            work_dir = self.builddir
            build_type_args = ["-o"]

            # go build ./... is not supported in go 1.11 or 1.12.
            # This will only install the main module.
            if self._get_parsed_go_version() < parse_version(
                _GO_MOD_ENV_FLAG_REQUIRED_GO_VERSION
            ):
                build_type_args.append(
                    os.path.join(self._install_bin_dir, self._get_module())
                )
            else:
                build_type_args.extend([self._install_bin_dir, "./..."])
        else:
            work_dir = self._install_bin_dir
            build_type_args = [package]

        pre_build_files = os.listdir(self._install_bin_dir)
        self._run(build_cmd + build_type_args, cwd=work_dir)
        post_build_files = os.listdir(self._install_bin_dir)

        new_files = set(post_build_files) - set(pre_build_files)

        if len(new_files) == 0:
            logger.warning(f"no binaries found from {build_cmd!r}")

        for new_file in new_files:
            binary_path = os.path.join(self._install_bin_dir, new_file)

            # Relink with system linker if executable is dynamic in order to be
            # able to set rpath later on. This workaround can be removed after
            # https://github.com/NixOS/patchelf/issues/146 is fixed.
            if self._is_classic and elf.ElfFile(path=binary_path).is_dynamic:
                self._run(relink_cmd + build_type_args, cwd=work_dir)

    def _build_go_packages(self) -> None:
        if self.options.go_packages:
            packages = self.options.go_packages
        else:
            packages = self._get_local_main_packages()

        for package in packages:
            self._build(package=package)

    def build(self) -> None:
        super().build()

        # Ensure install directory exists.
        os.makedirs(self._install_bin_dir, exist_ok=True)

        if self._is_using_go_mod(cwd=self.builddir):
            self._build()
        else:
            self._build_go_packages()

    def clean_build(self) -> None:
        super().clean_build()

        if os.path.isdir(self._gopath_bin):
            shutil.rmtree(self._gopath_bin)

        if os.path.isdir(self._gopath_pkg):
            shutil.rmtree(self._gopath_pkg)

    def _run(self, cmd: List[str], cwd: str = None, **kwargs) -> None:
        env = self._build_environment()

        if cwd is None:
            cwd = self._gopath_src

        return self.run(cmd, cwd=cwd, env=env, **kwargs)

    def _run_output(self, cmd: List[str], **kwargs) -> str:
        env = self._build_environment()
        return self.run_output(cmd, cwd=self._gopath_src, env=env, **kwargs)

    def _build_environment(self) -> Dict[str, str]:
        env = os.environ.copy()
        env["GOPATH"] = self._gopath
        env["GOBIN"] = self._gopath_bin

        library_paths: List[str] = []
        for root in [self.installdir, self.project.stage_dir]:
            library_paths.extend(
                common.get_library_paths(root, self.project.arch_triplet)
            )

        cgo_ldflags = _get_cgo_ldflags(library_paths)
        if cgo_ldflags:
            env["CGO_LDFLAGS"] = cgo_ldflags

        if self.project.is_cross_compiling:
            env["CC"] = "{}-gcc".format(self.project.arch_triplet)
            env["CXX"] = "{}-g++".format(self.project.arch_triplet)
            env["CGO_ENABLED"] = "1"
            # See https://golang.org/doc/install/source#environment
            go_archs = {"armhf": "arm", "i386": "386", "ppc64el": "ppc64le"}
            env["GOARCH"] = go_archs.get(self.project.deb_arch, self.project.deb_arch)
            if self.project.deb_arch == "armhf":
                env["GOARM"] = "7"
        return env

    def enable_cross_compilation(self) -> None:
        pass



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import io
import os

from testtools.matchers import Equals

from snapcraft import yaml_utils
from snapcraft.yaml_utils import YamlValidationError
from tests import unit


class YamlLoadTests(unit.TestCase):
    def test_load_yaml_file(self):
        path = os.path.join(self.path, "test.yaml")
        with open(path, "w") as f:
            f.write("foo: bar")

        cfg = yaml_utils.load_yaml_file(path)

        self.assertThat(cfg, Equals({"foo": "bar"}))

    def test_load_yaml_file_bad_yaml(self):
        path = os.path.join(self.path, "test.yaml")
        with open(path, "w") as f:
            # Note the missing newlines...
            f.write("foo: bar")
            f.write("a-b-c: xyz")

        self.assertRaises(YamlValidationError, yaml_utils.load_yaml_file, path)


class OctIntTest(unit.TestCase):
    def test_octint_dump(self):
        output = io.StringIO()
        yaml_utils.dump(dict(number=yaml_utils.OctInt(8)), stream=output)
        output.seek(0)

        self.assertThat(output.read().strip(), Equals("number: 0010"))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015, 2017-2018, 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

from testtools.matchers import Equals

import snapcraft
from snapcraft.plugins.v1.dump import DumpInvalidSymlinkError, DumpPlugin
from tests import unit


class DumpPluginTestCase(unit.TestCase):
    def setUp(self):
        super().setUp()
        self.project_options = snapcraft.ProjectOptions()

        class Options:
            source = "."

        self.options = Options()

    def test_dumping_nothing(self):
        plugin = DumpPlugin("dump", self.options, self.project_options)
        os.makedirs(plugin.builddir)
        plugin.build()

        self.assertThat(os.listdir(plugin.installdir), Equals([]))

    def test_dumping_with_contents(self):
        plugin = DumpPlugin("dump", self.options, self.project_options)

        os.makedirs(plugin.builddir)
        open(os.path.join(plugin.builddir, "file1"), "w").close()
        open(os.path.join(plugin.builddir, "file2"), "w").close()
        os.mkdir(os.path.join(plugin.builddir, "dir1"))
        open(os.path.join(plugin.builddir, "dir1", "subfile1"), "w").close()

        plugin.build()

        contents = os.listdir(plugin.installdir)
        contents.sort()
        self.assertThat(contents, Equals(["dir1", "file1", "file2"]))
        self.assertThat(
            os.listdir(os.path.join(plugin.installdir, "dir1")), Equals(["subfile1"])
        )

    def test_dump_symlinks(self):
        plugin = DumpPlugin("dump", self.options, self.project_options)

        os.makedirs(os.path.join(plugin.builddir, "subdir"))
        with open(os.path.join(plugin.builddir, "file"), "w") as f:
            f.write("foo")

        symlinks = [
            {
                "source": "file",
                "link_name": os.path.join(plugin.builddir, "relative1"),
                "destination": os.path.join(plugin.installdir, "relative1"),
                "expected_realpath": os.path.join(plugin.installdir, "file"),
                "expected_contents": "foo",
            },
            {
                "source": os.path.join("..", "file"),
                "link_name": os.path.join(plugin.builddir, "subdir", "relative2"),
                "destination": os.path.join(plugin.installdir, "subdir", "relative2"),
                "expected_realpath": os.path.join(plugin.installdir, "file"),
                "expected_contents": "foo",
            },
            {
                "source": os.path.join("..", "..", "install", "file"),
                "link_name": os.path.join(plugin.builddir, "subdir", "relative3"),
                "destination": os.path.join(plugin.installdir, "subdir", "relative3"),
                "expected_realpath": os.path.join(plugin.installdir, "file"),
                "expected_contents": "foo",
            },
        ]

        for symlink in symlinks:
            os.symlink(symlink["source"], symlink["link_name"])

        plugin.build()

        with open(os.path.join(plugin.installdir, "file"), "r") as f:
            self.assertThat(f.read(), Equals("foo"))

        for symlink in symlinks:
            destination = symlink["destination"]
            self.assertTrue(
                os.path.islink(destination),
                "Expected {!r} to be a symlink".format(destination),
            )

            self.assertThat(
                os.path.realpath(destination),
                Equals(symlink["expected_realpath"]),
                "Expected {!r} to be a relative path to {!r}".format(
                    destination, symlink["expected_realpath"]
                ),
            )

            with open(destination, "r") as f:
                self.assertThat(f.read(), Equals(symlink["expected_contents"]))

    def test_dump_symlinks_that_should_be_followed(self):
        # TODO: Move to an integration test
        plugin = DumpPlugin("dump", self.options, self.project_options)

        os.makedirs(os.path.join(plugin.builddir, "src"))
        with open(os.path.join(plugin.builddir, "src", "file"), "w") as f:
            f.write("foo")

        with open("unsnapped", "w") as f:
            f.write("bar")

        symlinks = [
            # Links with an absolute path should be followed
            {
                "source": os.path.abspath(os.path.join(plugin.builddir, "src", "file")),
                "link_name": os.path.join(plugin.builddir, "src", "absolute"),
                "destination": os.path.join(plugin.installdir, "src", "absolute"),
                "expected_contents": "foo",
            },
            # Links with a relative path that points outside of the snap
            # should also be followed
            {
                "source": "../../../../unsnapped",
                "link_name": os.path.join(plugin.builddir, "src", "bad_relative"),
                "destination": os.path.join(plugin.installdir, "src", "bad_relative"),
                "expected_contents": "bar",
            },
        ]

        for symlink in symlinks:
            os.symlink(symlink["source"], symlink["link_name"])

        plugin.build()

        with open(os.path.join(plugin.installdir, "src", "file"), "r") as f:
            self.assertThat(f.read(), Equals("foo"))

        for symlink in symlinks:
            destination = symlink["destination"]
            self.assertFalse(
                os.path.islink(destination),
                "Expected {!r} to be a copy rather than a "
                "symlink".format(destination),
            )

            with open(destination, "r") as f:
                self.assertThat(f.read(), Equals(symlink["expected_contents"]))

    def test_dump_symlinks_to_libc(self):
        plugin = DumpPlugin("dump", self.options, self.project_options)
        os.makedirs(plugin.builddir)

        # Even though this symlink is absolute, since it's to libc the copy
        # plugin shouldn't try to follow it or modify it.
        libc_libs = snapcraft.repo.Repo.get_package_libraries("libc6")

        # We don't care which lib we're testing with, as long as it's a .so.
        libc_library_path = [lib for lib in libc_libs if ".so" in lib][0]
        os.symlink(libc_library_path, os.path.join(plugin.builddir, "libc-link"))

        plugin.build()

        self.assertThat(
            os.path.join(plugin.installdir, "libc-link"),
            unit.LinkExists(libc_library_path),
        )

    def test_dump_broken_symlink(self):
        self.options.source = "src"
        plugin = DumpPlugin("dump", self.options, self.project_options)

        os.makedirs(os.path.join(plugin.builddir, "src"))
        with open(os.path.join(plugin.builddir, "src", "file"), "w") as f:
            f.write("foo")

        with open("unsnapped", "w") as f:
            f.write("bar")

        symlinks = [
            # This symlink is valid in source, but broken when snapped.
            {
                "source": "../../../unsnapped",
                "link_name": os.path.join(plugin.builddir, "src", "bad_relative"),
                "destination": os.path.join(plugin.installdir, "src", "bad_relative"),
                "expected_contents": "bar",
            }
        ]

        for symlink in symlinks:
            os.symlink(symlink["source"], symlink["link_name"])

        plugin.pull()

        raised = self.assertRaises(DumpInvalidSymlinkError, plugin.build)

        self.assertThat(
            raised.path, Equals(os.path.join(plugin.builddir, "src", "bad_relative"))
        )

    def test_dump_enable_cross_compilation(self):
        plugin = DumpPlugin("dump", self.options, self.project_options)
        plugin.enable_cross_compilation()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
from collections import OrderedDict
from copy import deepcopy
from typing import Any, Dict, List, Optional, Sequence, Set

from snapcraft import yaml_utils
from snapcraft.internal import common
from snapcraft.internal.meta import errors
from snapcraft.internal.meta.application import Application
from snapcraft.internal.meta.hooks import Hook
from snapcraft.internal.meta.package_repository import PackageRepository
from snapcraft.internal.meta.plugs import ContentPlug, Plug
from snapcraft.internal.meta.slots import ContentSlot, Slot
from snapcraft.internal.meta.system_user import SystemUser

logger = logging.getLogger(__name__)


class Snap:
    """Representation of snap meta object, writes snap.yaml."""

    def __init__(  # noqa: C901
        self,
        adopt_info: Optional[str] = None,
        apps: Optional[Dict[str, Application]] = None,
        architectures: Optional[Sequence[str]] = None,
        assumes: Optional[Set[str]] = None,
        base: Optional[str] = None,
        build_base: Optional[str] = None,
        compression: Optional[str] = None,
        confinement: Optional[str] = None,
        description: Optional[str] = None,
        environment: Optional[Dict[str, Any]] = None,
        epoch: Any = None,
        grade: Optional[str] = None,
        hooks: Optional[Dict[str, Hook]] = None,
        layout: Optional[Dict[str, Any]] = None,
        license: Optional[str] = None,
        name: Optional[str] = None,
        package_repositories: Optional[List[PackageRepository]] = None,
        passthrough: Optional[Dict[str, Any]] = None,
        plugs: Optional[Dict[str, Plug]] = None,
        slots: Optional[Dict[str, Slot]] = None,
        summary: Optional[str] = None,
        system_usernames: Optional[Dict[str, SystemUser]] = None,
        title: Optional[str] = None,
        type: Optional[str] = None,
        version: Optional[str] = None,
    ) -> None:
        self.adopt_info = adopt_info

        if apps is None:
            self.apps: Dict[str, Application] = dict()
        else:
            self.apps = apps

        if architectures is None:
            self.architectures: Sequence[str] = list()
        else:
            self.architectures = architectures

        if assumes is None:
            self.assumes: Set[str] = set()
        else:
            self.assumes = assumes

        self.base = base
        self.build_base = build_base
        self.compression = compression
        self.confinement = confinement
        self.description = description

        if environment is None:
            self.environment: Dict[str, Any] = dict()
        else:
            self.environment = environment

        self.epoch = epoch
        self.grade = grade

        if hooks is None:
            self.hooks: Dict[str, Hook] = dict()
        else:
            self.hooks = hooks

        if layout is None:
            self.layout: Dict[str, Any] = dict()
        else:
            self.layout = layout

        self.license = license
        self.name = name

        if package_repositories is None:
            self.package_repositories: List[PackageRepository] = list()
        else:
            self.package_repositories = package_repositories

        if passthrough is None:
            self.passthrough: Dict[str, Any] = dict()
        else:
            self.passthrough = passthrough

        if plugs is None:
            self.plugs: Dict[str, Plug] = dict()
        else:
            self.plugs = plugs

        if slots is None:
            self.slots: Dict[str, Slot] = dict()
        else:
            self.slots = slots

        self.summary = summary

        if system_usernames is None:
            self.system_usernames: Dict[str, SystemUser] = dict()
        else:
            self.system_usernames = system_usernames

        self.title = title
        self.type = type
        self.version = version

    @classmethod
    def from_file(cls, snap_yaml_path: str) -> "Snap":
        with open(snap_yaml_path, "r") as f:
            snap_dict = yaml_utils.load(f)
            return cls.from_dict(snap_dict=snap_dict)

    @property
    def is_passthrough_enabled(self) -> bool:
        if self.passthrough:
            return True

        for app in self.apps.values():
            if app.passthrough:
                return True

        for hook in self.hooks.values():
            if hook.passthrough:
                return True

        return False

    def get_build_base(self) -> str:
        """
        Return the base to use to create the snap.

        Returns build-base if set, but if not, name is returned if the
        snap is of type base. For all other snaps, the base is returned
        as the build-base.
        """
        build_base: Optional[str] = None
        if self.build_base is not None:
            build_base = self.build_base
        elif self.name is not None and self.type == "base":
            build_base = self.name
        else:
            build_base = self.base

        # The schema does not allow for this when loaded from snapcraft.yaml.
        if build_base is None:
            raise RuntimeError("'build_base' cannot be None")

        return build_base

    def get_content_plugs(self) -> List[ContentPlug]:
        """Get list of content plugs."""
        return [plug for plug in self.plugs.values() if isinstance(plug, ContentPlug)]

    def get_content_slots(self) -> List[ContentSlot]:
        """Get list of content slots."""
        return [slot for slot in self.slots.values() if isinstance(slot, ContentSlot)]

    def get_provider_content_directories(self) -> Set[str]:
        """Get provider content directories from installed snaps."""
        provider_dirs: Set[str] = set()

        for plug in self.get_content_plugs():
            # Get matching slot provider for plug.
            provider = plug.provider
            if not provider:
                continue

            provider_path = common.get_installed_snap_path(provider)
            yaml_path = os.path.join(provider_path, "meta", "snap.yaml")

            if not os.path.exists(yaml_path):
                continue

            snap = Snap.from_file(yaml_path)
            for slot in snap.get_content_slots():
                slot_installed_path = common.get_installed_snap_path(provider)
                provider_dirs |= slot.get_content_dirs(
                    installed_path=slot_installed_path
                )

        return provider_dirs

    def _validate_required_keys(self) -> None:
        """Verify that all mandatory keys have been satisfied."""
        missing_keys: List[str] = []

        if not self.name:
            missing_keys.append("name")

        if not self.version and not self.adopt_info:
            missing_keys.append("version")

        if not self.summary:
            missing_keys.append("summary")

        if not self.description:
            missing_keys.append("description")

        if missing_keys:
            raise errors.MissingSnapcraftYamlKeysError(keys=missing_keys)

    def validate(self) -> None:
        """Validate snap, raising exception on error."""
        self._validate_required_keys()

        for app in self.apps.values():
            app.validate()

        for hook in self.hooks.values():
            hook.validate()

        for plug in self.plugs.values():
            plug.validate()

        for slot in self.slots.values():
            slot.validate()

        for user in self.system_usernames.values():
            user.validate()

        if self.is_passthrough_enabled:
            logger.warning(
                "The 'passthrough' property is being used to "
                "propagate experimental properties to snap.yaml "
                "that have not been validated."
            )

    def _ensure_command_chain_assumption(self) -> None:
        """Ensure command-chain is in assumes (if used)."""
        if "command-chain" in self.assumes:
            return

        for app in self.apps.values():
            if app.command_chain:
                self.assumes.add("command-chain")
                return
        for hook in self.hooks.values():
            if hook.command_chain:
                self.assumes.add("command-chain")
                return

    @classmethod  # noqa: C901
    def from_dict(cls, snap_dict: Dict[str, Any]) -> "Snap":
        snap_dict = deepcopy(snap_dict)

        # Using pop() so we can catch if we *miss* fields
        # with whatever remains in the dictionary.
        adopt_info = snap_dict.pop("adopt-info", None)
        architectures = snap_dict.pop("architectures", None)

        # Process apps into Applications.
        apps: Dict[str, Application] = dict()
        apps_dict = snap_dict.pop("apps", None)
        if apps_dict:
            for app_name, app_dict in apps_dict.items():
                app = Application.from_dict(app_dict=app_dict, app_name=app_name)
                apps[app_name] = app

        # Treat `assumes` as a set, not as a list.
        assumes = set(snap_dict.pop("assumes", set()))

        base = snap_dict.pop("base", None)
        build_base = snap_dict.pop("build-base", None)
        compression = snap_dict.pop("compression", None)
        confinement = snap_dict.pop("confinement", None)
        description = snap_dict.pop("description", None)
        environment = snap_dict.pop("environment", None)
        epoch = snap_dict.pop("epoch", None)
        grade = snap_dict.pop("grade", None)

        # Process hooks into Hooks.
        hooks: Dict[str, Hook] = dict()
        hooks_dict = snap_dict.pop("hooks", None)
        if hooks_dict:
            for hook_name, hook_dict in hooks_dict.items():
                # This can happen, but should be moved into Hook.from_object().
                if hook_dict is None:
                    continue

                hook = Hook.from_dict(hook_dict=hook_dict, hook_name=hook_name)
                hooks[hook_name] = hook

        layout = snap_dict.pop("layout", None)
        license = snap_dict.pop("license", None)
        name = snap_dict.pop("name", None)

        raw_repositories = snap_dict.pop("package-repositories", None)
        if raw_repositories is None:
            package_repositories = None
        else:
            package_repositories = PackageRepository.unmarshal_package_repositories(
                raw_repositories
            )

        passthrough = snap_dict.pop("passthrough", None)

        # Process plugs into Plugs.
        plugs: Dict[str, Plug] = dict()
        plugs_dict = snap_dict.pop("plugs", None)
        if plugs_dict:
            for plug_name, plug_object in plugs_dict.items():
                plug = Plug.from_object(plug_object=plug_object, plug_name=plug_name)
                plugs[plug_name] = plug

        # Process slots into Slots.
        slots: Dict[str, Slot] = dict()
        slots_dict = snap_dict.pop("slots", None)
        if slots_dict:
            for slot_name, slot_object in slots_dict.items():
                slot = Slot.from_object(slot_object=slot_object, slot_name=slot_name)
                slots[slot_name] = slot

        summary = snap_dict.pop("summary", None)

        # Process sytemusers into SystemUsers.
        system_usernames: Dict[str, SystemUser] = dict()
        system_usernames_dict = snap_dict.pop("system-usernames", None)
        if system_usernames_dict:
            for user_name, user_object in system_usernames_dict.items():
                system_username = SystemUser.from_object(
                    user_object=user_object, user_name=user_name
                )
                system_usernames[user_name] = system_username

        title = snap_dict.pop("title", None)
        type = snap_dict.pop("type", None)
        version = snap_dict.pop("version", None)

        # Report unhandled keys.
        for key, value in snap_dict.items():
            logger.debug(f"ignoring or passing through unknown {key}={value}")

        return Snap(
            adopt_info=adopt_info,
            architectures=architectures,
            apps=apps,
            assumes=assumes,
            base=base,
            build_base=build_base,
            compression=compression,
            confinement=confinement,
            description=description,
            environment=environment,
            epoch=epoch,
            grade=grade,
            hooks=hooks,
            layout=layout,
            license=license,
            name=name,
            passthrough=passthrough,
            package_repositories=package_repositories,
            plugs=plugs,
            slots=slots,
            summary=summary,
            system_usernames=system_usernames,
            title=title,
            type=type,
            version=version,
        )

    def to_dict(self):  # noqa: C901
        snap_dict = OrderedDict()

        # Ensure command-chain is in assumes, if required.
        self._ensure_command_chain_assumption()

        if self.name is not None:
            snap_dict["name"] = self.name

        if self.version is not None:
            snap_dict["version"] = self.version

        if self.summary is not None:
            snap_dict["summary"] = self.summary

        if self.description is not None:
            snap_dict["description"] = self.description

        if self.adopt_info is not None:
            snap_dict["adopt-info"] = self.adopt_info

        if self.apps:
            snap_dict["apps"] = OrderedDict()
            for name, app in sorted(self.apps.items()):
                snap_dict["apps"][name] = deepcopy(app.to_dict())

        if self.architectures:
            snap_dict["architectures"] = deepcopy(self.architectures)

        if self.assumes:
            snap_dict["assumes"] = sorted(set(deepcopy(self.assumes)))

        if self.base is not None:
            snap_dict["base"] = self.base

        if self.build_base is not None:
            snap_dict["build-base"] = self.build_base

        if self.compression is not None:
            snap_dict["compression"] = self.compression

        if self.confinement is not None:
            snap_dict["confinement"] = self.confinement

        if self.environment:
            snap_dict["environment"] = self.environment

        if self.epoch is not None:
            snap_dict["epoch"] = self.epoch

        if self.grade is not None:
            snap_dict["grade"] = self.grade

        if self.hooks:
            snap_dict["hooks"] = OrderedDict()
            for name, hook in sorted(self.hooks.items()):
                snap_dict["hooks"][name] = deepcopy(hook.to_dict())

        if self.layout:
            snap_dict["layout"] = deepcopy(self.layout)

        if self.license is not None:
            snap_dict["license"] = self.license

        package_repos = [repo.marshal() for repo in self.package_repositories]
        if package_repos:
            snap_dict["package-repositories"] = package_repos

        if self.passthrough:
            snap_dict["passthrough"] = deepcopy(self.passthrough)

        if self.plugs:
            snap_dict["plugs"] = OrderedDict()
            for name, plug in sorted(self.plugs.items()):
                snap_dict["plugs"][name] = deepcopy(plug.to_yaml_object())

        if self.slots:
            snap_dict["slots"] = OrderedDict()
            for name, slot in sorted(self.slots.items()):
                snap_dict["slots"][name] = deepcopy(slot.to_yaml_object())

        if self.system_usernames:
            snap_dict["system-usernames"] = OrderedDict()
            for name, user in sorted(self.system_usernames.items()):
                snap_dict["system-usernames"][name] = deepcopy(user.to_dict())

        if self.title is not None:
            snap_dict["title"] = self.title

        if self.type is not None:
            snap_dict["type"] = self.type

        return snap_dict

    def to_snap_yaml_dict(self) -> OrderedDict:
        snap_dict = self.to_dict()

        # If the base is core in snapcraft.yaml we do not set it in
        # snap.yaml LP: #1819290
        if self.base == "core":
            snap_dict.pop("base")

        # Remove keys that are not for snap.yaml.
        snap_dict.pop("build-base", None)
        snap_dict.pop("adopt-info", None)
        snap_dict.pop("compression", None)
        snap_dict.pop("package-repositories", None)

        # Apply passthrough keys.
        passthrough = snap_dict.pop("passthrough", dict())
        snap_dict.update(passthrough)

        return snap_dict

    def write_snap_yaml(self, path: str) -> None:
        """Write snap.yaml contents to specified path."""
        snap_dict = self.to_snap_yaml_dict()

        with open(path, "w") as f:
            yaml_utils.dump(snap_dict, stream=f, sort_keys=False)

    def __repr__(self) -> str:
        return repr(self.__dict__)

    def __str__(self) -> str:
        return str(self.__dict__)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import snapcraft.formatting_utils
import snapcraft.internal.errors


class PythonPluginError(snapcraft.internal.errors.SnapcraftError):
    pass


class MissingPythonCommandError(PythonPluginError):

    fmt = "Unable to find {python_version}, searched: {search_paths}"

    def __init__(self, python_version, search_paths):
        super().__init__(
            python_version=python_version,
            search_paths=snapcraft.formatting_utils.combine_paths(
                search_paths, "", ":"
            ),
        )


class MissingUserSitePackagesError(PythonPluginError):

    fmt = "Unable to find user site packages: {site_dir_glob}"

    def __init__(self, site_dir_glob):
        super().__init__(site_dir_glob=site_dir_glob)


class MissingSitePyError(PythonPluginError):

    fmt = "Unable to find site.py: {site_py_glob}"

    def __init__(self, site_py_glob):
        super().__init__(site_py_glob=site_py_glob)


class PipListInvalidLegacyFormatError(PythonPluginError):

    fmt = (
        "Failed to parse Python package list: "
        "The returned output is not in the expected format:\n"
        "{output}"
    )

    def __init__(self, output):
        super().__init__(output=output)


class PipListInvalidJsonError(PythonPluginError):

    fmt = "Pip packages output isn't valid json: {json!r}"

    def __init__(self, json):
        super().__init__(json=json)


class PipListMissingFieldError(PythonPluginError):

    fmt = "Pip packages json missing {field!r} field: {json!r}"

    def __init__(self, field, json):
        super().__init__(field=field, json=json)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
from collections import OrderedDict
from typing import Dict, List

from snapcraft import formatting_utils
from snapcraft.internal.errors import SnapcraftError

_VALIDATION_ERROR_CAUSES = {
    "maxLength": "maximum length is {validator_value}",
    "minLength": "minimum length is {validator_value}",
}


class YamlValidationError(SnapcraftError):

    fmt = "Issues while validating {source}: {message}"

    @classmethod
    def from_validation_error(cls, error, *, source="snapcraft.yaml"):
        """Take a jsonschema.ValidationError and create a SnapcraftSchemaError.

        The validation errors coming from jsonschema are a nightmare. This
        class tries to make them a bit more understandable.
        """

        messages = []

        preamble = _determine_preamble(error)
        cause = _determine_cause(error)
        supplement = _determine_supplemental_info(error)

        if preamble:
            messages.append(preamble)

        # If we have a preamble we are not at the root
        if supplement and preamble:
            messages.append(error.message)
            messages.append("({})".format(supplement))
        elif supplement:
            messages.append(supplement)
        elif cause:
            messages.append(cause)
        else:
            messages.append(error.message)

        return cls(" ".join(messages), source)

    def __init__(self, message, source="snapcraft.yaml"):
        super().__init__(message=message, source=source)


def _determine_preamble(error):
    messages = []
    path = _determine_property_path(error)
    if path:
        messages.append(
            "The '{}' property does not match the required schema:".format(
                "/".join(path)
            )
        )
    return " ".join(messages)


def _determine_cause(error):
    messages = []

    # error.validator_value may contain a custom validation error message.
    # If so, use it instead of the garbage message jsonschema gives us.
    with contextlib.suppress(TypeError, KeyError):
        messages.append(error.validator_value["validation-failure"].format(error))

    # The schema itself may have a custom validation error message. If so,
    # use it as well.
    with contextlib.suppress(AttributeError, TypeError, KeyError):
        key = error
        if (
            error.schema.get("type") == "object"
            and error.validator == "additionalProperties"
        ):
            key = list(error.instance.keys())[0]

        messages.append(error.schema["validation-failure"].format(key))

    # anyOf failures might have usable context... try to improve them a bit
    if error.validator == "anyOf":
        contextual_messages: Dict[str, str] = OrderedDict()
        for contextual_error in error.context:
            key = contextual_error.schema_path.popleft()
            if key not in contextual_messages:
                contextual_messages[key] = []
            message = contextual_error.message
            if message:
                # Sure it starts lower-case (not all messages do)
                contextual_messages[key].append(message[0].lower() + message[1:])

        oneOf_messages: List[str] = []
        for key, value in contextual_messages.items():
            oneOf_messages.append(formatting_utils.humanize_list(value, "and", "{}"))

        messages.append(formatting_utils.humanize_list(oneOf_messages, "or", "{}"))

    return " ".join(messages)


def _determine_supplemental_info(error):
    message = _VALIDATION_ERROR_CAUSES.get(error.validator, "").format(
        validator_value=error.validator_value
    )

    if not message and error.validator == "anyOf":
        message = _interpret_anyOf(error)

    if not message and error.cause:
        message = error.cause

    return message


def _determine_property_path(error):
    path = []
    while error.absolute_path:
        element = error.absolute_path.popleft()
        # assume numbers are indices and use 'xxx[123]' notation.
        if isinstance(element, int):
            path[-1] = "{}[{}]".format(path[-1], element)
        else:
            path.append(str(element))

    return path


def _interpret_anyOf(error):
    """Interpret a validation error caused by the anyOf validator.

    Returns:
        A string containing a (hopefully) helpful validation error message. It
        may be empty.
    """

    usages = []
    try:
        for validator in error.validator_value:
            usages.append(validator["usage"])
    except (TypeError, KeyError):
        return ""

    return "must be one of {}".format(
        formatting_utils.humanize_list(usages, "or", "{}")
    )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from typing import Optional
from unittest import mock

from testtools.matchers import Contains, Equals

from snapcraft.internal import steps

from . import LifecycleCommandsBaseTestCase


class TestSnap(LifecycleCommandsBaseTestCase):
    def assert_build_provider_calls(
        self, shell: bool = False, output: Optional[str] = None
    ):
        self.fake_lifecycle_execute.mock.assert_not_called()

        if shell:
            self.provider_mock.shell.assert_called_once_with()
        else:
            self.provider_mock.pack_project.assert_called_once_with(output=output)
            self.provider_mock.shell.assert_not_called()

        self.assert_clean_not_called()

    def test_using_defaults(self):
        result = self.run_command(["snap"])

        self.assertThat(result.exit_code, Equals(0))
        self.fake_get_provider_for.mock.assert_called_once_with("multipass")
        self.assert_build_provider_calls()

    def test_output_using_defaults(self):
        result = self.run_command(["snap", "--output", "foo.snap"])

        self.assertThat(result.exit_code, Equals(0))
        self.fake_get_provider_for.mock.assert_called_once_with("multipass")
        self.assert_build_provider_calls(output="foo.snap")

    def test_output_using_defaults_with_dir(self):
        result = self.run_command(["snap", "--output", "/tmp"])

        self.assertThat(result.exit_code, Equals(0))
        self.fake_get_provider_for.mock.assert_called_once_with("multipass")
        self.assert_build_provider_calls(output="/tmp")

    def test_shell_using_defaults(self):
        result = self.run_command(["snap", "--shell"])

        self.assertThat(result.exit_code, Equals(0))
        self.fake_get_provider_for.mock.assert_called_once_with("multipass")
        self.assert_build_provider_calls(shell=True)
        self.provider_mock.execute_step.assert_called_once_with(steps.PRIME)

    def test_shell_after_using_defaults(self):
        result = self.run_command(["snap", "--shell-after"])

        self.assertThat(result.exit_code, Equals(0))
        self.fake_get_provider_for.mock.assert_called_once_with("multipass")
        self.assert_build_provider_calls(shell=True)
        self.provider_mock.execute_step.assert_not_called()

    def test_using_lxd(self):
        result = self.run_command(["snap", "--use-lxd"])

        self.assertThat(result.exit_code, Equals(0))
        self.fake_get_provider_for.mock.assert_called_once_with("lxd")
        self.assert_build_provider_calls()

    def test_using_destructive_mode(self):
        result = self.run_command(["snap", "--destructive-mode"])

        self.assertThat(result.exit_code, Equals(0))
        self.fake_get_provider_for.mock.assert_not_called()
        self.fake_lifecycle_execute.mock.assert_called_once_with(
            steps.PRIME, mock.ANY, tuple()
        )
        self.fake_pack.mock.assert_called_once_with(
            os.path.join(self.path, "prime"), compression=None, output=None
        )

    def test_output_using_destructive_mode(self):
        result = self.run_command(
            ["snap", "--destructive-mode", "--output", "foo.snap"]
        )

        self.assertThat(result.exit_code, Equals(0))
        self.fake_get_provider_for.mock.assert_not_called()
        self.fake_lifecycle_execute.mock.assert_called_once_with(
            steps.PRIME, mock.ANY, tuple()
        )
        self.fake_pack.mock.assert_called_once_with(
            os.path.join(self.path, "prime"), compression=None, output="foo.snap"
        )

    def test_output_using_destructive_mode_with_directory(self):
        result = self.run_command(["snap", "--destructive-mode", "--output", "/tmp"])

        self.assertThat(result.exit_code, Equals(0))
        self.fake_get_provider_for.mock.assert_not_called()
        self.fake_lifecycle_execute.mock.assert_called_once_with(
            steps.PRIME, mock.ANY, tuple()
        )
        self.fake_pack.mock.assert_called_once_with(
            os.path.join(self.path, "prime"), compression=None, output="/tmp"
        )

    def test_deprecated_snap_dir(self):
        result = self.run_command(["snap", "snap-dir"])

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(result.output, Contains("DEPRECATED"))
        self.fake_get_provider_for.mock.assert_not_called()
        self.fake_lifecycle_execute.mock.assert_not_called()
        self.fake_pack.mock.assert_called_once_with("snap-dir", output=None)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import shutil
import subprocess
from textwrap import dedent

from testtools.matchers import Equals, FileExists

from tests import integration


class GitSourceTestCase(integration.GitSourceBaseTestCase):
    def _get_git_revno(self, path, revrange="-1"):
        return subprocess.check_output(
            "git -C {} log {} --oneline | cut -d' ' -f2".format(path, revrange),
            shell=True,
            universal_newlines=True,
        ).strip()

    def test_pull_git_head(self):
        self.copy_project_to_cwd("git-head")

        self.init_source_control()
        self.commit('"1"', allow_empty=True)
        self.commit('"2"', allow_empty=True)

        self.run_snapcraft("pull")
        revno = self._get_git_revno("parts/git/src")
        self.assertThat(revno, Equals('"2"'))

        self.run_snapcraft("pull")
        revno = self._get_git_revno("parts/git/src")
        self.assertThat(revno, Equals('"2"'))

    def test_pull_git_tag(self):
        self.copy_project_to_cwd("git-tag")

        self.init_source_control()
        self.commit('"1"', allow_empty=True)
        self.commit('"2"', allow_empty=True)
        subprocess.check_call(
            ["git", "tag", "initial", "HEAD@{1}"], stdout=subprocess.DEVNULL
        )

        self.run_snapcraft("pull")
        revno = self._get_git_revno("parts/git/src")
        self.assertThat(revno, Equals('"1"'))

        self.run_snapcraft("pull")
        revno = self._get_git_revno("parts/git/src")
        self.assertThat(revno, Equals('"1"'))

    def test_pull_git_commit(self):
        self.copy_project_to_cwd("git-commit")

        self.init_source_control()
        self.commit('"1"', allow_empty=True)
        self.commit('"2"', allow_empty=True)

        # The test uses "HEAD^" so we can only test it once
        self.run_snapcraft("pull")
        revno = self._get_git_revno("parts/git/src")
        self.assertThat(revno, Equals('"1"'))

    def test_pull_git_branch(self):
        self.copy_project_to_cwd("git-branch")

        self.init_source_control()
        self.commit('"1"', allow_empty=True)
        self.commit('"2"', allow_empty=True)
        subprocess.check_call(
            ["git", "branch", "second", "HEAD@{1}"], stdout=subprocess.DEVNULL
        )
        subprocess.check_call(["git", "checkout", "second"], stderr=subprocess.DEVNULL)
        self.commit('"3"', allow_empty=True)
        subprocess.check_call(["git", "checkout", "master"], stderr=subprocess.DEVNULL)

        self.run_snapcraft("pull")
        revno = self._get_git_revno("parts/git/src", revrange="-2")
        self.assertThat(revno, Equals('"3"\n"1"'))

        self.run_snapcraft("pull")
        revno = self._get_git_revno("parts/git/src", revrange="-2")
        self.assertThat(revno, Equals('"3"\n"1"'))

    def test_pull_git_with_depth(self):
        """Regression test for LP: #1627772."""
        self.copy_project_to_cwd("git-depth")

        self.init_source_control()
        self.commit('"1"', allow_empty=True)
        self.commit('"2"', allow_empty=True)

        self.run_snapcraft("pull")


class GitGenerateVersionTestCase(integration.GitSourceBaseTestCase):
    def setUp(self):
        super().setUp()
        self.init_source_control()
        os.mkdir("snap")

        with open(os.path.join("snap", "snapcraft.yaml"), "w") as f:
            print(
                dedent(
                    """\
                name: git-test
                version: git
                summary: test git generated version
                description: test git generated version with git hint
                parts:
                    nil:
                        plugin: nil
                """
                ),
                file=f,
            )

        self.add_file(os.path.join("snap", "snapcraft.yaml"))
        self.commit("snapcraft.yaml added")

    def test_tag(self):
        self.tag("2.0")
        self.run_snapcraft("snap")
        self.assertThat("git-test_2.0_{}.snap".format(self.deb_arch), FileExists())

    def test_tag_with_commits_ahead(self):
        self.tag("2.0")
        open("stub_file", "w").close()
        self.add_file("stub_file")
        self.commit("new stub file")
        self.run_snapcraft("snap")
        revno = self.get_revno()[:7]
        expected_file = "git-test_2.0+git1.{}_{}.snap".format(revno, self.deb_arch)
        self.assertThat(expected_file, FileExists())

    def test_no_tag(self):
        self.run_snapcraft("snap")
        revno = self.get_revno()[:7]
        expected_file = "git-test_0+git.{}_{}.snap".format(revno, self.deb_arch)
        self.assertThat(expected_file, FileExists())

    def test_no_git(self):
        shutil.rmtree(".git")

        # Do not check the output as it is dependent on the tool used (git)
        # and it has changed across releases.
        self.assertRaises(subprocess.CalledProcessError, self.run_snapcraft, ["snap"])



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020-2021 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The cmake plugin is useful for building cmake based parts.

These are projects that have a CMakeLists.txt that drives the build.
The plugin requires a CMakeLists.txt in the root of the source tree.

This plugin uses the common plugin keywords as well as those for "sources".
For more information check the 'plugins' topic for the former and the
'sources' topic for the latter.

Additionally, this plugin uses the following plugin-specific keywords:

    - cmake-parameters
      (list of strings)
      parameters to pass to the build using the common cmake semantics.

    - cmake-generator
      (string; default: "Unix Makefiles")
      Determine what native build system is to be used.
      Can be either `ninja` or `Unix Makefiles` (default).

This plugin also interprets these specific build-environment entries:

    - SNAPCRAFT_CMAKE_ARGS
      Additional arguments to pass to the cmake configuration.
"""

from typing import Any, Dict, List, Set

from snapcraft.plugins.v2 import PluginV2


class CMakePlugin(PluginV2):
    @classmethod
    def get_schema(cls) -> Dict[str, Any]:
        return {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "type": "object",
            "additionalProperties": False,
            "properties": {
                "cmake-parameters": {
                    "type": "array",
                    "uniqueItems": True,
                    "items": {"type": "string"},
                    "default": [],
                },
                "cmake-generator": {
                    "type": "string",
                    "enum": ["Unix Makefiles", "Ninja"],
                    "default": "Unix Makefiles",
                },
            },
        }

    def get_build_snaps(self) -> Set[str]:
        return set()

    def get_build_packages(self) -> Set[str]:
        build_packages = {"gcc", "cmake"}

        if self.options.cmake_generator == "Ninja":
            build_packages.add("ninja-build")

        return build_packages

    def get_build_environment(self) -> Dict[str, str]:
        return {
            "CMAKE_PREFIX_PATH": "${SNAPCRAFT_STAGE}",
            "SNAPCRAFT_CMAKE_ARGS": "",
        }

    def _get_cmake_configure_command(self) -> str:
        cmd = [
            "cmake",
            '"${SNAPCRAFT_PART_SRC_WORK}"',
            "-G",
            f'"{self.options.cmake_generator}"',
            "${SNAPCRAFT_CMAKE_ARGS}",
        ] + self.options.cmake_parameters

        return " ".join(cmd)

    @property
    def out_of_source_build(self):
        return True

    def get_build_commands(self) -> List[str]:
        return [
            self._get_cmake_configure_command(),
            'cmake --build . -- -j"${SNAPCRAFT_PARALLEL_BUILD_COUNT}"',
            'DESTDIR="${SNAPCRAFT_PART_INSTALL}" cmake --build . --target install',
        ]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import pathlib
import subprocess
from unittest import mock

from testtools.matchers import Contains, Equals

import snapcraft
from snapcraft.plugins.v1._ros import wstool
from tests import unit


class WstoolTestCase(unit.TestCase):
    def setUp(self):
        super().setUp()
        self.project = snapcraft.ProjectOptions()
        self.wstool = wstool.Wstool("package_path", "wstool_path", self.project, "core")

        patcher = mock.patch("snapcraft.repo.Ubuntu")
        self.ubuntu_mock = patcher.start()
        self.addCleanup(patcher.stop)

        patcher = mock.patch("subprocess.check_output")
        self.check_output_mock = patcher.start()
        self.addCleanup(patcher.stop)

    def test_setup(self):
        # Return something other than a Mock to ease later assertions
        self.check_output_mock.return_value = b""

        self.wstool.setup()

        # Verify that only wstool was installed (no other .debs)
        self.assertThat(
            self.ubuntu_mock.fetch_stage_packages.mock_calls,
            Equals(
                [
                    mock.call(
                        stage_packages_path=self.wstool._wstool_stage_packages_path,
                        package_names=["python-wstool"],
                        base="core",
                        target_arch=self.project.target_arch,
                    )
                ]
            ),
        )
        self.assertThat(
            self.ubuntu_mock.unpack_stage_packages.mock_calls,
            Equals(
                [
                    mock.call(
                        stage_packages_path=self.wstool._wstool_stage_packages_path,
                        install_path=pathlib.Path("wstool_path/install"),
                    )
                ]
            ),
        )

        # Verify that wstool was initialized
        self.check_output_mock.assert_called_once_with(
            ["wstool", "init", "package_path", "-j2"],
            env=mock.ANY,
            stderr=subprocess.PIPE,
        )

    def test_setup_can_run_multiple_times(self):
        self.wstool.setup()

        # Make sure running setup() again doesn't have problems with the old
        # environment. An exception will be raised if setup can't be called
        # twice.
        self.wstool.setup()

    def test_setup_initialization_rosinstall_already_exists(self):
        """Test that an existing .rosinstall file is not an error."""

        def run(args, **kwargs):
            if args[0:2] == ["wstool", "init"]:
                raise subprocess.CalledProcessError(
                    1,
                    "foo",
                    b"Error: There already is a workspace config file "
                    b'.rosinstall at ".". Use wstool install/modify.',
                )

        self.check_output_mock.side_effect = run

        self.wstool.setup()

    def test_setup_initialization_failure(self):
        def run(args, **kwargs):
            if args[0:2] == ["wstool", "init"]:
                raise subprocess.CalledProcessError(1, "foo", b"bar", b"baz")

        self.check_output_mock.side_effect = run

        raised = self.assertRaises(
            wstool.WorkspaceInitializationError, self.wstool.setup
        )

        self.assertThat(str(raised), Equals("Error initializing workspace: baz"))

    def test_merge(self):
        self.wstool.merge("rosinstall-file")

        self.check_output_mock.assert_called_with(
            ["wstool", "merge", "rosinstall-file", "--confirm-all", "-tpackage_path"],
            stderr=subprocess.PIPE,
            env=mock.ANY,
        )

    def test_merge_failure(self):
        def run(args, **kwargs):
            if args[0:2] == ["wstool", "merge"]:
                raise subprocess.CalledProcessError(1, "foo", b"bar", b"baz")

        self.check_output_mock.side_effect = run

        raised = self.assertRaises(
            wstool.RosinstallMergeError, self.wstool.merge, "rosinstall-file"
        )

        self.assertThat(
            str(raised),
            Equals(
                "Error merging rosinstall file 'rosinstall-file' into workspace: baz"
            ),
        )

    def test_update(self):
        self.wstool.update()

        self.check_output_mock.assert_called_with(
            ["wstool", "update", "-j2", "-tpackage_path"],
            env=mock.ANY,
            stderr=subprocess.PIPE,
        )

    def test_update_failure(self):
        def run(args, **kwargs):
            if args[0:2] == ["wstool", "update"]:
                raise subprocess.CalledProcessError(1, "foo", b"bar", b"baz")

        self.check_output_mock.side_effect = run

        raised = self.assertRaises(wstool.WorkspaceUpdateError, self.wstool.update)

        self.assertThat(str(raised), Equals("Error updating workspace: baz"))

    def test_run(self):
        wstool = self.wstool

        # Create a single library directory for asserting LD_LIBRARY_PATH is set
        # properly.
        os.makedirs(os.path.join(wstool._wstool_install_path, "lib"))

        wstool._run(["init"])

        class check_env:
            def __init__(self, test):
                self.test = test

            def __eq__(self, env):
                for variable in (
                    "PATH",
                    "LD_LIBRARY_PATH",
                    "GIT_EXEC_PATH",
                    "PYTHONPATH",
                ):
                    self.test.assertThat(env, Contains(variable))

                self.test.assertThat(
                    env["PATH"],
                    Equals(
                        "{}:{}".format(
                            os.environ["PATH"],
                            os.path.join(wstool._wstool_install_path, "usr", "bin"),
                        )
                    ),
                )

                self.test.assertThat(
                    env["LD_LIBRARY_PATH"],
                    Equals(
                        "{}".format(os.path.join(wstool._wstool_install_path, "lib"))
                    ),
                )

                self.test.assertThat(
                    env["GIT_EXEC_PATH"],
                    Equals(
                        os.path.join(
                            wstool._wstool_install_path, "usr", "lib", "git-core"
                        )
                    ),
                )

                self.test.assertThat(
                    env["PYTHONPATH"],
                    Equals(
                        os.path.join(
                            wstool._wstool_install_path,
                            "usr",
                            "lib",
                            "python2.7",
                            "dist-packages",
                        )
                    ),
                )

                return True

        self.check_output_mock.assert_called_with(
            mock.ANY, env=check_env(self), stderr=subprocess.PIPE
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os
from textwrap import dedent
from unittest import mock

from testtools.matchers import Equals

from tests.unit import TestWithFakeRemoteParts

from . import CommandBaseTestCase


class RefreshCommandBaseTestCase(CommandBaseTestCase, TestWithFakeRemoteParts):

    yaml_template = dedent(
        """\
        name: snap-test
        version: "1.0"
        summary: test snapping
        description: if snap is successful a snap package will be available
        architectures: ['amd64']
        type: {}
        confinement: strict
        grade: stable

        parts:
            part1:
                plugin: nil
        """
    )

    def make_snapcraft_yaml(self, n=1, snap_type="app", snapcraft_yaml=None):
        if not snapcraft_yaml:
            snapcraft_yaml = self.yaml_template.format(snap_type)
        super().make_snapcraft_yaml(snapcraft_yaml)
        self.state_dir = os.path.join(self.parts_dir, "part1", "state")


class RefreshCommandTestCase(RefreshCommandBaseTestCase):
    @mock.patch("snapcraft.cli.containers.repo.Repo.refresh_build_packages")
    def test_refresh(self, mock_repo_refresh):
        self.make_snapcraft_yaml()

        result = self.run_command(["refresh"])

        self.assertThat(result.exit_code, Equals(0))
        mock_repo_refresh.assert_called_once_with()
